# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.core as jc
import jax.tree_util as jtu
from jax import util as jax_util
from jax.extend import source_info_util as src_util

from genjax._src.core.interpreters.forward import (
    Environment,
    InitialStylePrimitive,
    initial_style_bind,
)
from genjax._src.core.interpreters.staging import stage
from genjax._src.core.pytree import Closure, Pytree
from genjax._src.core.typing import (
    Any,
    ArrayLike,
    Callable,
    Generic,
    TypeVar,
)

R = TypeVar("R")
S = TypeVar("S")

record_p = InitialStylePrimitive("record_p")


@Pytree.dataclass
class FrameRecording(Generic[R, S], Pytree):
    f: Callable[..., R]
    args: tuple[Any, ...]
    local_retval: R
    cont: Callable[..., S]


@Pytree.dataclass
class RecordPoint(Generic[R, S], Pytree):
    callable: Closure[R]
    debug_tag: str | None = Pytree.static()

    def default_call(self, *args) -> R:
        return self.callable(*args)

    def handle(self, cont: Callable[[R], tuple[S, Any]], *args):
        @Pytree.partial()
        def _cont(*args) -> S:
            final_ret, _ = cont(self.callable(*args))
            return final_ret

        # Normal execution.
        ret = self.callable(*args)
        final_ret = _cont(*args)
        return final_ret, (
            self.debug_tag,
            FrameRecording(self.callable, args, ret, _cont),
        )

    def __call__(self, *args):
        def _cont_prim_call(brk_pt, *args):
            return brk_pt.default_call(*args)

        return initial_style_bind(record_p)(_cont_prim_call)(self, *args)


def rec(
    callable: Callable[..., R],
    debug_tag: str | None = None,
):
    if not isinstance(callable, Closure):
        callable = Closure[R]((), callable)

    def inner(*args):
        return RecordPoint(callable, debug_tag)(*args)

    return inner


def tag(v, name=None):
    return rec(lambda v: v, name)(v)


##########################
# Hybrid CPS interpreter #
##########################


@Pytree.dataclass
class TimeTravelCPSInterpreter(Pytree):
    @staticmethod
    def _eval_jaxpr_hybrid_cps(
        jaxpr: jc.Jaxpr,
        consts: list[ArrayLike],
        flat_args: list[ArrayLike],
        out_tree,
    ):
        env = Environment()
        jax_util.safe_map(env.write, jaxpr.constvars, consts)
        jax_util.safe_map(env.write, jaxpr.invars, flat_args)

        # Hybrid CPS evaluation.
        def eval_jaxpr_iterate_cps(
            eqns,
            env: Environment,
            invars,
            flat_args,
            rebind=False,
        ):
            jax_util.safe_map(env.write, invars, flat_args)

            for eqn_idx, eqn in enumerate(eqns):
                with src_util.user_context(eqn.source_info.traceback):
                    invals = jax_util.safe_map(env.read, eqn.invars)
                    subfuns, params = eqn.primitive.get_bind_params(eqn.params)
                    args = subfuns + invals

                    if eqn.primitive == record_p:
                        env = env.copy()

                        @Pytree.partial()
                        def _kont(*args):
                            leaves = jtu.tree_leaves(args)
                            return eval_jaxpr_iterate_cps(
                                eqns[eqn_idx + 1 :],
                                env,
                                eqn.outvars,
                                leaves,
                                rebind=True,
                            )

                        in_tree = params["in_tree"]
                        num_consts = params["num_consts"]
                        cps_prim, *args = jtu.tree_unflatten(in_tree, args[num_consts:])
                        if rebind:
                            return _kont(cps_prim(*args))

                        else:
                            return cps_prim.handle(_kont, *args)

                    else:
                        outs = eqn.primitive.bind(*args, **params)

                if not eqn.primitive.multiple_results:
                    outs = [outs]

                jax_util.safe_map(
                    env.write,
                    eqn.outvars,
                    outs,
                )

            out_values = jax.util.safe_map(
                env.read,
                jaxpr.outvars,
            )
            retval = jtu.tree_unflatten(out_tree(), out_values)
            return retval, None

        return eval_jaxpr_iterate_cps(
            jaxpr.eqns,
            env,
            jaxpr.invars,
            flat_args,
        )

    @staticmethod
    def time_travel(f):
        def _inner(*args):
            closed_jaxpr, (flat_args, _, out_tree) = stage(f)(*args)
            jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.literals
            return TimeTravelCPSInterpreter._eval_jaxpr_hybrid_cps(
                jaxpr,
                consts,
                flat_args,
                out_tree,
            )

        return _inner


def time_travel(f):
    return TimeTravelCPSInterpreter.time_travel(f)


@Pytree.dataclass
class TimeTravelingDebugger(Pytree):
    final_retval: Any
    sequence: list[FrameRecording[Any, Any]]
    jump_points: dict[Any, Any] = Pytree.static()
    ptr: int = Pytree.static()

    def frame(self) -> tuple[str | None, FrameRecording[Any, Any]]:
        frame = self.sequence[self.ptr]
        reverse_jump_points = {v: k for (k, v) in self.jump_points.items()}
        jump_tag = reverse_jump_points.get(self.ptr, None)
        return jump_tag, frame

    def summary(self) -> tuple[Any, tuple[str | None, FrameRecording[Any, Any]]]:
        frame = self.sequence[self.ptr]
        reverse_jump_points = {v: k for (k, v) in self.jump_points.items()}
        jump_tag = reverse_jump_points.get(self.ptr, None)
        return self.final_retval, (jump_tag, frame)

    def jump(self, debug_tag: str) -> "TimeTravelingDebugger":
        jump_pt = self.jump_points[debug_tag]
        return TimeTravelingDebugger(
            self.final_retval,
            self.sequence,
            self.jump_points,
            jump_pt,
        )

    def fwd(self) -> "TimeTravelingDebugger":
        new_ptr = self.ptr + 1
        if new_ptr >= len(self.sequence):
            return self
        else:
            return TimeTravelingDebugger(
                self.final_retval,
                self.sequence,
                self.jump_points,
                self.ptr + 1,
            )

    def bwd(self) -> "TimeTravelingDebugger":
        new_ptr = self.ptr - 1
        if new_ptr >= len(self.sequence) or new_ptr < 0:
            return self
        else:
            return TimeTravelingDebugger(
                self.final_retval,
                self.sequence,
                self.jump_points,
                new_ptr,
            )

    def remix(self, *args) -> "TimeTravelingDebugger":
        frame = self.sequence[self.ptr]
        f, cont = frame.f, frame.cont
        local_retval = f(*args)
        _, debugger = _record(cont)(*args)
        new_frame = FrameRecording(f, args, local_retval, cont)
        return TimeTravelingDebugger(
            debugger.final_retval,
            [*self.sequence[: self.ptr], new_frame, *debugger.sequence],
            self.jump_points,
            self.ptr,
        )

    def __call__(self, *args):
        return self.remix(*args)


def _record(source: Callable[..., Any]):
    def inner(*args) -> tuple[Any, TimeTravelingDebugger]:
        retval, next = time_travel(source)(*args)
        sequence = []
        jump_points = {}
        while next:
            (debug_tag, frame) = next
            sequence.append(frame)
            if debug_tag:
                jump_points[debug_tag] = len(sequence) - 1
            args, cont = frame.args, frame.cont
            retval, next = time_travel(cont)(*args)
        return retval, TimeTravelingDebugger(retval, sequence, jump_points, 0)

    return inner


def time_machine(source: Callable[..., Any]):
    def instrumented(*args):
        return tag(rec(source, "_enter")(*args), "exit")

    def inner(*args) -> TimeTravelingDebugger:
        _, debugger = _record(instrumented)(*args)
        return debugger

    return inner



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The `combinators` module exposes _generative function combinators_: generative
functions which accept other generative functions as configuration arguments, and
implement their own generative function interfaces using structured patterns of control
flow (and other types of useful modifications). If one thinks of a control flow
primitive as an operation on deterministic types, a combinator can be thought of as
lifting the operation to support generative function semantics.

GenJAX exposes several combinators:

* [`MaskedCombinator`](masked.md) - which can mask a generative computation based on a runtime determined `BoolArray` argument.
* [`VmapCombinator`](map.md) - which exposes generative vectorization over input arguments. The implementation essentially wraps [`jax.vmap`](https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html) into the interfaces.
* [`ScanCombinator`](unfold.md) - which exposes a scan-like pattern for generative computation in a state space pattern, by utilizing the control flow primitive [`jax.lax.scan`](https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html).
* [`SwitchCombinator`](switch.md) - which exposes stochastic branching patterns, by utilizing the control flow primitive [`jax.lax.switch`](https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.switch.html).
"""



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains an abstract data class (called `Pytree`) for implementing JAX's [`Pytree` interface](https://jax.readthedocs.io/en/latest/pytrees.html) on derived classes.

The Pytree interface determines how data classes behave across JAX-transformed function boundaries - it provides a user with the freedom to declare subfields of a class as "static" (meaning, the value of the field cannot be a JAX traced value, it must be a Python literal, or a constant array - and the value is embedded in the `PyTreeDef` of any instance) or "dynamic" (meaning, the value may be a JAX traced value).
"""

from dataclasses import field
from typing import overload

import jax.numpy as jnp
import jax.tree_util as jtu
import penzai.pz as pz
import treescope
from treescope import formatting_util
from typing_extensions import dataclass_transform

from genjax._src.core.typing import (
    Any,
    Callable,
    Generic,
    TypeVar,
    static_check_is_concrete,
)

R = TypeVar("R")


class Pytree(pz.Struct):
    """`Pytree` is an abstract base class which registers a class with JAX's `Pytree`
    system. JAX's `Pytree` system tracks how data classes should behave across JAX-transformed function boundaries, like `jax.jit` or `jax.vmap`.

    Inheriting this class provides the implementor with the freedom to declare how the subfields of a class should behave:

    * `Pytree.static(...)`: the value of the field cannot be a JAX traced value, it must be a Python literal, or a constant). The values of static fields are embedded in the `PyTreeDef` of any instance of the class.
    * `Pytree.field(...)` or no annotation: the value may be a JAX traced value, and JAX will attempt to convert it to tracer values inside of its transformations.

    If a field _points to another `Pytree`_, it should not be declared as `Pytree.static()`, as the `Pytree` interface will automatically handle the `Pytree` fields as dynamic fields.

    """

    @staticmethod
    @overload
    def dataclass(
        incoming: None = None,
        /,
        **kwargs,
    ) -> Callable[[type[R]], type[R]]: ...

    @staticmethod
    @overload
    def dataclass(
        incoming: type[R],
        /,
        **kwargs,
    ) -> type[R]: ...

    @dataclass_transform(
        frozen_default=True,
    )
    @staticmethod
    def dataclass(
        incoming: type[R] | None = None,
        /,
        **kwargs,
    ) -> type[R] | Callable[[type[R]], type[R]]:
        """
        Denote that a class (which is inheriting `Pytree`) should be treated as a dataclass, meaning it can hold data in fields which are declared as part of the class.

        A dataclass is to be distinguished from a "methods only" `Pytree` class, which does not have fields, but may define methods.
        The latter cannot be instantiated, but can be inherited from, while the former can be instantiated:
        the `Pytree.dataclass` declaration informs the system _how to instantiate_ the class as a dataclass,
        and how to automatically define JAX's `Pytree` interfaces (`tree_flatten`, `tree_unflatten`, etc.) for the dataclass, based on the fields declared in the class, and possibly `Pytree.static(...)` or `Pytree.field(...)` annotations (or lack thereof, the default is that all fields are `Pytree.field(...)`).

        All `Pytree` dataclasses support pretty printing, as well as rendering to HTML.

        Examples:
            ```python exec="yes" html="true" source="material-block" session="core"
            from genjax import Pytree
            from genjax.typing import FloatArray
            import jax.numpy as jnp


            @Pytree.dataclass
            # Enforces type annotations on instantiation.
            class MyClass(Pytree):
                my_static_field: int = Pytree.static()
                my_dynamic_field: FloatArray


            print(MyClass(10, jnp.array(5.0)).render_html())
            ```
        """

        return pz.pytree_dataclass(
            incoming,
            overwrite_parent_init=True,
            **kwargs,
        )

    @staticmethod
    def static(**kwargs):
        """Declare a field of a `Pytree` dataclass to be static. Users can provide additional keyword argument options,
        like `default` or `default_factory`, to customize how the field is instantiated when an instance of
        the dataclass is instantiated.` Fields which are provided with default values must come after required fields in the dataclass declaration.

        Examples:
            ```python exec="yes" html="true" source="material-block" session="core"
            @Pytree.dataclass
            # Enforces type annotations on instantiation.
            class MyClass(Pytree):
                my_dynamic_field: FloatArray
                my_static_field: int = Pytree.static(default=0)


            print(MyClass(jnp.array(5.0)).render_html())
            ```

        """
        return field(metadata={"pytree_node": False}, **kwargs)

    @staticmethod
    def field(**kwargs):
        "Declare a field of a `Pytree` dataclass to be dynamic. Alternatively, one can leave the annotation off in the declaration."
        return field(**kwargs)

    ##############################
    # Utility class constructors #
    ##############################

    @staticmethod
    def const(v):
        # The value must be concrete!
        # It cannot be a JAX traced value.
        assert static_check_is_concrete(v)
        if isinstance(v, Const):
            return v
        else:
            return Const(v)

    # Safe: will not wrap a Const in another Const, and will not
    # wrap dynamic values.
    @staticmethod
    def tree_const(v):
        def _inner(v):
            if isinstance(v, Const):
                return v
            elif static_check_is_concrete(v):
                return Const(v)
            else:
                return v

        return jtu.tree_map(
            _inner,
            v,
            is_leaf=lambda v: isinstance(v, Const),
        )

    @staticmethod
    def tree_const_unwrap(v):
        def _inner(v):
            if isinstance(v, Const):
                return v.val
            else:
                return v

        return jtu.tree_map(
            _inner,
            v,
            is_leaf=lambda v: isinstance(v, Const),
        )

    @staticmethod
    def partial(*args) -> Callable[[Callable[..., R]], "Closure[R]"]:
        return lambda fn: Closure[R](args, fn)

    def treedef(self):
        return jtu.tree_structure(self)

    #################
    # Static checks #
    #################

    @staticmethod
    def static_check_tree_structure_equivalence(trees: list[Any]):
        if not trees:
            return True
        else:
            fst, *rest = trees
            treedef = jtu.tree_structure(fst)
            check = all(map(lambda v: treedef == jtu.tree_structure(v), rest))
            return check

    def treescope_color(self) -> str:
        """Computes a CSS color to display for this object in treescope.

        This function can be overridden to change the color for a particular object
        in treescope, without having to register a new handler.

        (note that we are overriding the Penzai base class's implementation so that ALL structs receive colors, not just classes with `__call__` implemented.)

        Returns:
          A CSS color string to use as a background/highlight color for this object.
          Alternatively, a tuple of (border, fill) CSS colors.
        """
        type_string = type(self).__module__ + "." + type(self).__qualname__
        return formatting_util.color_from_string(type_string)

    def render_html(self):
        return treescope.render_to_html(
            self,
            roundtrip_mode=False,
        )


##############################
# Associated utility classes #
##############################


# Wrapper for static values (can include callables).
@Pytree.dataclass
class Const(Generic[R], Pytree):
    """
    JAX-compatible way to tag a value as a constant. Valid constants include Python literals, strings, essentially anything **that won't hold JAX arrays** inside of a computation.

    Examples:
        Instances of `Const` can be created using a `Pytree` classmethod:
        ```python exec="yes" html="true" source="material-block" session="core"
        from genjax import Pytree

        c = Pytree.const(5)
        print(c.render_html())
        ```

        Constants can be freely used across [`jax.jit`](https://jax.readthedocs.io/en/latest/_autosummary/jax.jit.html) boundaries:
        ```python exec="yes" html="true" source="material-block" session="core"
        from genjax import Pytree


        def f(c):
            if c.const == 5:
                return 10.0
            else:
                return 5.0


        c = Pytree.const(5)
        r = jax.jit(f)(c)
        print(r)
        ```
    """

    val: R = Pytree.static()

    def __call__(self, *args):
        assert isinstance(self.val, Callable), (
            f"Wrapped `val` {self.val} is not Callable."
        )
        return self.val(*args)

    def unwrap(self: Any) -> R:
        """Unwrap a constant value from a `Const` instance.

        This method can be used as an instance method or as a static method. When used as a static method, it returns the input value unchanged if it is not a `Const` instance.

        Returns:
            R: The unwrapped value if self is a `Const`, otherwise returns self unchanged.

        Examples:
            ```python exec="yes" html="true" source="material-block" session="core"
            from genjax import Pytree, Const

            c = Pytree.const(5)
            val = c.unwrap()  # Returns 5

            # Can also be used as static method
            val = Const.unwrap(10)  # Returns 10 unchanged
            ```
        """
        if isinstance(self, Const):
            return self.val
        else:
            return self


# Construct for a type of closure which closes over dynamic values.
@Pytree.dataclass
class Closure(Generic[R], Pytree):
    """
    JAX-compatible closure type. It's a closure _as a [`Pytree`][genjax.core.Pytree]_ - meaning the static _source code_ / _callable_ is separated from dynamic data (which must be tracked by JAX).

    Examples:
        Instances of `Closure` can be created using `Pytree.partial` -- note the order of the "closed over" arguments:
        ```python exec="yes" html="true" source="material-block" session="core"
        from genjax import Pytree


        def g(y):
            @Pytree.partial(y)  # dynamic values come first
            def f(v, x):
                # v will be bound to the value of y
                return x * (v * 5.0)

            return f


        clos = jax.jit(g)(5.0)
        print(clos.render_html())
        ```

        Closures can be invoked / JIT compiled in other code:
        ```python exec="yes" html="true" source="material-block" session="core"
        r = jax.jit(lambda x: clos(x))(3.0)
        print(r)
        ```
    """

    dyn_args: tuple[Any, ...]
    fn: Callable[..., R] = Pytree.static()

    def __call__(self, *args, **kwargs) -> R:
        return self.fn(*self.dyn_args, *args, **kwargs)


def nth(x: Pytree, idx: int | slice | jnp.ndarray):
    """Returns a Pytree in which `[idx]` has been applied to every leaf."""
    return jtu.tree_map(lambda v: v[idx], x)


class PythonicPytree(Pytree):
    """
    A class that adds support for bracket indexing/slicing, sequence-like operations,
    and concatenation to make working with Pytrees more Pythonic. The base class is
    appropriate for Pytrees which have a unform shape over leaves (or at least each
    leaf's initial axis should have the same length).
    """

    def __getitem__(self, idx):
        """Return a pytree in which each leaf has been sliced by `idx`."""
        return nth(self, idx)

    def __len__(self):
        """Return the "length" of the Pytree. This should only be used on
        Pytrees which have a uniform shape over leaves; it operates by
        returning the length of the "first" leaf."""
        return len(jtu.tree_leaves(self)[0])

    def __iter__(self):
        """Returs an iterator which generates each self[i] in order"""
        return (self[i] for i in range(len(self)))

    def __add__(self, other):
        """Concatenates two pytrees, leaf-wise."""
        if not isinstance(other, type(self)):
            raise TypeError(f"Cannot add {type(self)} and {type(other)}")

        def concat_leaves(x, y):
            return jnp.concatenate([x, y])

        return jtu.tree_map(concat_leaves, self, other)

    def prepend(self, child):
        """Prepends a scalar element to the front of each leaf in a Pytree."""
        return jtu.tree_map(lambda x: x[jnp.newaxis], child) + self



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import functools

import jax.numpy as jnp
import jax.tree_util as jtu
from jax.experimental import checkify

from genjax._src.checkify import optional_check
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.interpreters.staging import FlagOp, tree_choose
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Array,
    ArrayLike,
    Flag,
    Generic,
    TypeVar,
)

R = TypeVar("R")


#########################
# Masking and sum types #
#########################


@Pytree.dataclass(match_args=True, init=False)
class Mask(Generic[R], Pytree):
    """The `Mask` datatype wraps a value in a Boolean flag which denotes whether the data is valid or invalid to use in inference computations.

    Masks can be used in a variety of ways as part of generative computations - their primary role is to denote data which is valid under inference computations. Valid data can be used as `ChoiceMap` leaves, and participate in generative and inference computations (like scores, and importance weights or density ratios). A Mask with a False flag **should** be considered unusable, and should be handled with care.

    If a `flag` has a non-scalar shape, that implies that the mask is vectorized, and that the `ArrayLike` value, or each leaf in the pytree, must have the flag's shape as its prefix (i.e., must have been created with a `jax.vmap` call or via a GenJAX `vmap` combinator).

    ## Encountering `Mask` in your computation

    When users see `Mask` in their computations, they are expected to interact with them by either:

    * Unmasking them using the `Mask.unmask` interface, a potentially unsafe operation.

    * Destructuring them manually, and handling the cases.

    ## Usage of invalid data

    If you use invalid `Mask(data, False)` data in inference computations, you may encounter silently incorrect results.
    """

    value: R
    flag: Flag | Diff[Flag]

    ################
    # Constructors #
    ################

    def __init__(self, v: R, f: Flag | Diff[Flag] = True) -> None:
        assert not isinstance(v, Mask), (
            f"Mask should not be instantiated with another Mask! found {v}"
        )
        Mask._validate_init(v, f)

        self.value, self.flag = v, f  # pyright: ignore[reportAttributeAccessIssue]

    @staticmethod
    def _validate_init(value: R, flag: Flag | Diff[Flag]) -> None:
        """Validates that non-scalar flags are only used with vectorized masks.

        When a flag has a non-scalar shape (e.g. shape (3,)), this indicates the mask is vectorized.
        In this case, each leaf value in the pytree must have the flag's shape as a prefix of its own shape.
        For example, if flag has shape (3,), then array leaves must have shapes like (3,), (3,4), (3,2,1) etc.

        This ensures that vectorized flags properly align with vectorized data.

        Args:
            value: The value to be masked, can be a pytree
            flag: The flag to apply, either a scalar or array flag

        Raises:
            ValueError: If a non-scalar flag's shape is not a prefix of all leaf value shapes
        """
        flag = flag.get_primal() if isinstance(flag, Diff) else flag
        f_shape = jnp.shape(flag)
        if f_shape == ():
            return None

        leaf_shapes = [jnp.shape(leaf) for leaf in jtu.tree_leaves(value)]
        prefix_len = len(f_shape)

        for shape in leaf_shapes:
            if shape[:prefix_len] != f_shape:
                raise ValueError(
                    f"Vectorized flag {flag}'s shape {f_shape} must be a prefix of all leaf shapes. Found {shape}."
                )

    @staticmethod
    def _validate_leaf_shapes(this: R, other: R):
        """Validates that two values have matching shapes at each leaf.

        Used by __or__, __xor__ etc. to ensure we only combine masks with values whose leaves have matching shapes.
        Broadcasting is not supported - array shapes must match exactly.

        Args:
            this: First value to compare
            other: Second value to compare

        Raises:
            ValueError: If any leaf shapes don't match exactly
        """

        # Check array shapes match exactly (no broadcasting)
        def check_leaf_shapes(x, y):
            x_shape = jnp.shape(x)
            y_shape = jnp.shape(y)
            if x_shape != y_shape:
                raise ValueError(
                    f"Cannot combine masks with different array shapes: {x_shape} vs {y_shape}"
                )
            return None

        jtu.tree_map(check_leaf_shapes, this, other)

    def _validate_mask_shapes(self, other: "Mask[R]") -> None:
        """Used by __or__, __xor__ etc. to ensure we only combine masks with matching pytree shape and matching leaf shapes."""
        if jtu.tree_structure(self.value) != jtu.tree_structure(other.value):
            raise ValueError("Cannot combine masks with different tree structures!")

        Mask._validate_leaf_shapes(self, other)
        return None

    @staticmethod
    def build(v: "R | Mask[R]", f: Flag | Diff[Flag] = True) -> "Mask[R]":
        """
        Create a Mask instance, potentially from an existing Mask or a raw value.

        This method allows for the creation of a new Mask or the modification of an existing one. If the input is already a Mask, it combines the new flag with the existing one using a logical AND operation.

        Args:
            v: The value to be masked. Can be a raw value or an existing Mask.
            f: The flag to be applied to the value.

        Returns:
            A new Mask instance with the given value and flag.

        Note:
            If `v` is already a Mask, the new flag is combined with the existing one using a logical AND, ensuring that the resulting Mask is only valid if both input flags are valid.
        """
        match v:
            case Mask(value, g):
                assert not isinstance(f, Diff) and not isinstance(g, Diff)
                assert FlagOp.is_scalar(f) or (jnp.shape(f) == jnp.shape(g)), (
                    f"Can't build a Mask with non-matching Flag shapes {jnp.shape(f)} and {jnp.shape(g)}"
                )
                return Mask[R](value, FlagOp.and_(f, g))
            case _:
                return Mask[R](v, f)

    @staticmethod
    def maybe_mask(v: "R | Mask[R]", f: Flag) -> "R | Mask[R] | None":
        """
        Create a Mask instance or return the original value based on the flag.

        This method is similar to `build`, but it handles concrete flag values differently. For concrete True flags, it returns the original value without wrapping it in a Mask. For concrete False flags, it returns None. For non-concrete flags, it creates a new Mask instance.

        Args:
            v: The value to be potentially masked. Can be a raw value or an existing Mask.
            f: The flag to be applied to the value.

        Returns:
            - The original value `v` if `f` is concretely True.
            - None if `f` is concretely False.
            - A new Mask instance with the given value and flag if `f` is not concrete.
        """
        return Mask.build(v, f).flatten()

    #############
    # Accessors #
    #############

    def __getitem__(self, path) -> "Mask[R]":
        path = path if isinstance(path, tuple) else (path,)

        f = self.primal_flag()
        if isinstance(f, Array) and f.shape:
            # A non-scalar flag must have been produced via vectorization. Because a scalar flag can
            # wrap a non-scalar value, only use the vectorized components of the path to index into the flag...
            f = f[path[: len(f.shape)]]

        # but the use full path to index into the value.
        v_idx = jtu.tree_map(lambda v: v[path], self.value)

        # Reconstruct Diff if needed
        if isinstance(self.flag, Diff):
            f = Diff(f, self.flag.tangent)

        return Mask.build(v_idx, f)

    def flatten(self) -> "R | Mask[R] | None":
        """
        Flatten a Mask instance into its underlying value or None.

        "Flattening" occurs when the flag value is a concrete Boolean (True/False). In these cases, the Mask is simplified to either its raw value or None. If the flag is not concrete (i.e., a symbolic/traced value), the Mask remains intact.

        This method evaluates the mask's flag and returns:
        - None if the flag is concretely False or the value is None
        - The raw value if the flag is concretely True
        - The Mask instance itself if the flag is not concrete

        Returns:
            The flattened result based on the mask's flag state.
        """
        flag = self.primal_flag()
        if FlagOp.concrete_false(flag):
            return None
        elif FlagOp.concrete_true(flag):
            return self.value
        else:
            return self

    def unmask(self, default: R | None = None) -> R:
        """
        Unmask the `Mask`, returning the value within.

        This operation is inherently unsafe with respect to inference semantics if no default value is provided. It is only valid if the `Mask` wraps valid data at runtime, or if a default value is supplied.

        Args:
            default: An optional default value to return if the mask is invalid.

        Returns:
            The unmasked value if valid, or the default value if provided and the mask is invalid.
        """
        if default is None:

            def _check():
                checkify.check(
                    jnp.all(self.primal_flag()),
                    "Attempted to unmask when a mask flag (or some flag in a vectorized mask) is False: the unmasked value is invalid.\n",
                )

            optional_check(_check)
            return self.value
        else:

            def inner(true_v: ArrayLike, false_v: ArrayLike) -> Array:
                return jnp.where(self.primal_flag(), true_v, false_v)

            return jtu.tree_map(inner, self.value, default)

    def primal_flag(self) -> Flag:
        """
        Returns the primal flag of the mask.

        This method retrieves the primal (non-`Diff`-wrapped) flag value. If the flag
        is a Diff type (which contains both primal and tangent components), it returns
        the primal component. Otherwise, it returns the flag as is.

        Returns:
            The primal flag value.
        """
        match self.flag:
            case Diff(primal, _):
                return primal
            case flag:
                return flag

    ###############
    # Combinators #
    ###############

    def _or_idx(self, first: Flag, second: Flag):
        """Converts a pair of flag arrays into an array of indices for selecting between two values.

        This function implements a truth table for selecting between two values based on their flags:

        first | second | output | meaning
        ------+--------+--------+------------------
            0   |   0    |   -1   | neither valid
            1   |   0    |    0   | first valid only
            0   |   1    |    1   | second valid only
            1   |   1    |    0   | both valid for OR, invalid for XOR

        The output index is used to select between the corresponding values:
           0 -> select first value
           1 -> select second value

        Args:
            first: The flag for the first value
            second: The flag for the second value

        Returns:
            An Array of indices (-1, 0, or 1) indicating which value to select from each side.
        """
        # Note that the validation has already run to check that these flags have the same shape.
        return first + 2 * FlagOp.and_(FlagOp.not_(first), second) - 1

    def __or__(self, other: "Mask[R]") -> "Mask[R]":
        self._validate_mask_shapes(other)

        match self.primal_flag(), other.primal_flag():
            case True, _:
                return self
            case False, _:
                return other
            case self_flag, other_flag:
                idx = self._or_idx(self_flag, other_flag)
                return tree_choose(idx, [self, other])

    def __xor__(self, other: "Mask[R]") -> "Mask[R]":
        self._validate_mask_shapes(other)

        match self.primal_flag(), other.primal_flag():
            case (False, False) | (True, True):
                return Mask.build(self, False)
            case True, False:
                return self
            case False, True:
                return other
            case self_flag, other_flag:
                idx = self._or_idx(self_flag, other_flag)

                # note that `idx` above will choose the correct side for the FF, FT and TF cases,
                # but will equal 0 for TT flags. We use `FlagOp.xor_` to override this flag to equal
                # False, since neither side in the TT case will provide a `False` flag for us.
                chosen = tree_choose(idx, [self.value, other.value])
                return Mask(chosen, FlagOp.xor_(self_flag, other_flag))

    def __invert__(self) -> "Mask[R]":
        not_flag = jtu.tree_map(FlagOp.not_, self.flag)
        return Mask(self.value, not_flag)

    @staticmethod
    def or_n(mask: "Mask[R]", *masks: "Mask[R]") -> "Mask[R]":
        """Performs an n-ary OR operation on a sequence of Mask objects.

        Args:
            mask: The first mask to combine
            *masks: Variable number of additional masks to combine with OR

        Returns:
            A new Mask combining all inputs with OR operations
        """
        return functools.reduce(lambda a, b: a | b, masks, mask)

    @staticmethod
    def xor_n(mask: "Mask[R]", *masks: "Mask[R]") -> "Mask[R]":
        """Performs an n-ary XOR operation on a sequence of Mask objects.

        Args:
            mask: The first mask to combine
            *masks: Variable number of additional masks to combine with XOR

        Returns:
            A new Mask combining all inputs with XOR operations
        """
        return functools.reduce(lambda a, b: a ^ b, masks, mask)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides the core functionality and JAX compatibility layer which several
`GenJAX` generative function and inference modules are built on top of.

It contains (truncated, and in no particular order):

* Core data types for the associated data types of generative functions.

* Utility abstract data types (mixins) for automatically registering class definitions as valid `Pytree` method implementors (guaranteeing `flatten`/`unflatten` compatibility across JAX transform boundaries). For more information, see [Pytrees](https://jax.readthedocs.io/en/latest/pytrees.html).

* Transformation interpreters: interpreter-based transformations which operate on `ClosedJaxpr` instances, as well as staging functionality for staging out computations to `ClosedJaxpr` instances. The core interpreters are written in a mixed initial / final style. The application of all interpreters are JAX compatible, meaning that the application of any interpreter can be staged out to eliminate the interpreter overhead.
"""



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from functools import wraps

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from jax import core as jc
from jax import util as jax_util
from jax.extend import linear_util as lu
from jax.extend import source_info_util as src_util
from jax.interpreters import ad as jax_autodiff
from jax.interpreters import batching

from genjax._src.core.interpreters.forward import (
    Environment,
    InitialStylePrimitive,
    batch_fun,
    initial_style_bind,
)
from genjax._src.core.interpreters.staging import stage
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Annotated,
    Any,
    ArrayLike,
    Callable,
    Is,
    PRNGKey,
)

DualTree = Annotated[
    Any,
    Is[lambda v: Dual.static_check_dual_tree(v)],
]
"""
`DualTree` is the type of `Pytree` argument values with `Dual` leaves.
"""

###################
# ADEV primitives #
###################


class ADEVPrimitive(Pytree):
    """
    An `ADEVPrimitive` is a primitive sampler equipped with a JVP gradient estimator strategy. These objects support forward sampling, but also come equipped with a strategy that interacts with ADEV's AD transformation to return a JVP estimate.
    """

    @abstractmethod
    def sample(self, key, *args) -> Any:
        pass

    @abstractmethod
    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,  # Pytree with Dual leaves.
        konts: tuple[Callable[..., Any], Callable[..., Any]],
    ) -> "Dual":
        pass

    def get_batched_prim(self, dims: tuple[Any, ...]) -> "ADEVPrimitive":
        """
        To use ADEV primitives inside of `vmap`, they must provide a custom batched primitive version of themselves.

        This method returns the batched primitive, which contains customized gradient estimator strategies which are compatible with batching.
        """
        raise NotImplementedError

    def __call__(self, *args):
        return sample_primitive(self, *args)


class TailCallADEVPrimitive(ADEVPrimitive):
    @abstractmethod
    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,  # Pytree with Dual leaves.
    ) -> "Dual":
        pass

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,  # Pytree with Dual leaves.
        konts: tuple[Callable[..., Any], Callable[..., Any]],
    ) -> "Dual":
        _, kdual = konts
        return kdual(key, self.before_tail_call(key, dual_tree))

    def get_batched_prim(self, dims: tuple[Any, ...]):
        return TailCallBatchedADEVPrimitive(self, dims)


@Pytree.dataclass
class TailCallBatchedADEVPrimitive(TailCallADEVPrimitive):
    original_prim: TailCallADEVPrimitive
    dims: tuple[Any, ...] = Pytree.static()

    def sample(self, key, *args):
        return jax.vmap(self.original_prim.sample, in_axes=self.dims)(key, *args)

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,  # Pytree with Dual leaves.
    ) -> "Dual":
        key_dim, tree_dim = jax_util.split_list(self.dims, [1])
        tree_primals = Dual.tree_primal(dual_tree)
        tree_tangents = Dual.tree_tangent(dual_tree)

        def _before_tail_call(key, tree_primals, tree_tangents):
            dual_tree = Dual.dual_tree(tree_primals, tree_tangents)
            return self.original_prim.before_tail_call(key, dual_tree)

        return jax.vmap(
            _before_tail_call,
            in_axes=(key_dim, tree_dim, tree_dim),
        )(key, tree_primals, tree_tangents)


####################
# Sample intrinsic #
####################


sample_p = InitialStylePrimitive("sample")


def sample_primitive(adev_prim: ADEVPrimitive, *args, key=jax.random.key(0)):
    def _adev_prim_call(adev_prim, *args):
        # When used for abstract tracing, value of the key doesn't matter.
        # However, we support overloading the key for other transformations,
        # which will rely on the default semantics of `initial_style_bind`,
        # which is to call this function -- then, the value of the key will matter.
        v = adev_prim.sample(key, *args)
        return v

    return initial_style_bind(sample_p)(_adev_prim_call)(
        adev_prim,
        *args,
    )


# TODO: this is gnarly as fuck.
def batch_primitive(args, dims, **params):
    def fun_impl(*args, **params):
        consts, args = jax_util.split_list(args, [params["num_consts"]])
        return jc.eval_jaxpr(params["_jaxpr"], consts, *args)

    batched, out_dims = batch_fun(lu.wrap_init(fun_impl, params), dims)

    # populate the out_dims generator
    _ = batched.call_wrapped(*args)  # pyright: ignore

    # Now, we construct our actual batch primitive, and insert it
    # into the IR by binding it via `sample_primitive`.
    in_tree = params["in_tree"]
    key, *rest = args
    adev_prim, *primals = jtu.tree_unflatten(in_tree, rest)
    batched_prim = adev_prim.get_batched_prim(dims)

    # Insert into the IR.
    v = sample_primitive(
        batched_prim,
        *primals,
        key=key,
    )

    # TODO: static check on out_dims?
    return jtu.tree_leaves(v), out_dims()


batching.primitive_batchers[sample_p] = batch_primitive


####################
# ADEV interpreter #
####################


@Pytree.dataclass
class Dual(Pytree):
    primal: Any
    tangent: Any

    @staticmethod
    def tree_pure(v):
        def _inner(v):
            if isinstance(v, Dual):
                return v
            else:
                return Dual(v, jnp.zeros_like(v))

        return jtu.tree_map(_inner, v, is_leaf=lambda v: isinstance(v, Dual))

    @staticmethod
    def dual_tree(primals, tangents):
        return jtu.tree_map(lambda v1, v2: Dual(v1, v2), primals, tangents)

    @staticmethod
    def tree_primal(v):
        def _inner(v):
            if isinstance(v, Dual):
                return v.primal
            else:
                return v

        return jtu.tree_map(_inner, v, is_leaf=lambda v: isinstance(v, Dual))

    @staticmethod
    def tree_tangent(v):
        def _inner(v):
            if isinstance(v, Dual):
                return v.tangent
            else:
                return v

        return jtu.tree_map(_inner, v, is_leaf=lambda v: isinstance(v, Dual))

    @staticmethod
    def tree_leaves(v):
        v = Dual.tree_pure(v)
        return jtu.tree_leaves(v, is_leaf=lambda v: isinstance(v, Dual))

    @staticmethod
    def tree_unzip(v):
        primals = jtu.tree_leaves(Dual.tree_primal(v))
        tangents = jtu.tree_leaves(Dual.tree_tangent(v))
        return tuple(primals), tuple(tangents)

    @staticmethod
    def static_check_is_dual(v) -> bool:
        return isinstance(v, Dual)

    @staticmethod
    def static_check_dual_tree(v) -> bool:
        return all(
            map(
                lambda v: isinstance(v, Dual),
                jtu.tree_leaves(v, is_leaf=Dual.static_check_is_dual),
            )
        )


@Pytree.dataclass
class ADInterpreter(Pytree):
    """The `ADInterpreter` takes a `Jaxpr`,
    propagates dual numbers through it, while also performing a CPS transformation,
    to compute forward mode AD.

    When this interpreter hits
    the `sample_p` primitive, it creates a pair of continuation closures which is passed to the gradient strategy which the primitive is using.
    """

    @staticmethod
    def flat_unzip(duals: list[Any]):
        primals, tangents = jax_util.unzip2((t.primal, t.tangent) for t in duals)
        return list(primals), list(tangents)

    @staticmethod
    def _eval_jaxpr_adev_jvp(
        key: PRNGKey,
        jaxpr: jc.Jaxpr,
        consts: list[ArrayLike],
        flat_duals: list[Dual],
    ):
        dual_env = Environment()
        jax_util.safe_map(dual_env.write, jaxpr.constvars, Dual.tree_pure(consts))
        jax_util.safe_map(dual_env.write, jaxpr.invars, flat_duals)

        # TODO: Pure evaluation.
        def eval_jaxpr_iterate_pure(key, eqns, pure_env, invars, flat_args):
            jax_util.safe_map(pure_env.write, invars, flat_args)
            for eqn in eqns:
                in_vals = jax_util.safe_map(pure_env.read, eqn.invars)
                subfuns, params = eqn.primitive.get_bind_params(eqn.params)
                args = subfuns + in_vals
                if eqn.primitive is sample_p:
                    pass
                else:
                    outs = eqn.primitive.bind(*args, **params)
                    if not eqn.primitive.multiple_results:
                        outs = [outs]
                    jax_util.safe_map(pure_env.write, eqn.outvars, outs)

            return jax_util.safe_map(pure_env.read, jaxpr.outvars)

        # Dual evaluation.
        def eval_jaxpr_iterate_dual(key, eqns, dual_env, invars, flat_duals):
            jax_util.safe_map(dual_env.write, invars, flat_duals)

            for eqn_idx, eqn in enumerate(eqns):
                with src_util.user_context(eqn.source_info.traceback):
                    in_vals = jax_util.safe_map(dual_env.read, eqn.invars)
                    subfuns, params = eqn.primitive.get_bind_params(eqn.params)
                    duals = subfuns + in_vals

                    # Our sample_p primitive.
                    if eqn.primitive is sample_p:
                        dual_env = dual_env.copy()
                        pure_env = Dual.tree_primal(dual_env)

                        # Create pure continuation.
                        def _sample_pure_kont(key, *args):
                            return eval_jaxpr_iterate_pure(
                                key,
                                eqns[eqn_idx + 1 :],
                                pure_env,
                                eqn.outvars,
                                [*args],
                            )

                        # Create dual continuation.
                        def _sample_dual_kont(key, dual_tree):
                            dual_leaves = Dual.tree_leaves(dual_tree)
                            return eval_jaxpr_iterate_dual(
                                key,
                                eqns[eqn_idx + 1 :],
                                dual_env,
                                eqn.outvars,
                                dual_leaves,
                            )

                        in_tree = params["in_tree"]
                        num_consts = params["num_consts"]

                        flat_primals, flat_tangents = ADInterpreter.flat_unzip(
                            Dual.tree_leaves(Dual.tree_pure(duals[num_consts:]))
                        )
                        adev_prim, *primals = jtu.tree_unflatten(in_tree, flat_primals)
                        _, *tangents = jtu.tree_unflatten(in_tree, flat_tangents)
                        dual_tree = Dual.dual_tree(primals, tangents)

                        return adev_prim.jvp_estimate(
                            key,
                            dual_tree,
                            (_sample_pure_kont, _sample_dual_kont),
                        )

                    # Handle branching.
                    elif eqn.primitive is jax.lax.cond_p:
                        pure_env = Dual.tree_primal(dual_env)

                        # Create dual continuation for the computation after the cond_p.
                        def _cond_dual_kont(dual_tree: list[Any]):
                            dual_leaves = Dual.tree_pure(dual_tree)
                            return eval_jaxpr_iterate_dual(
                                key,
                                eqns[eqn_idx + 1 :],
                                dual_env,
                                eqn.outvars,
                                dual_leaves,
                            )

                        branch_adev_functions = list(
                            map(
                                lambda fn: ADInterpreter.forward_mode(
                                    jc.jaxpr_as_fun(fn),
                                    _cond_dual_kont,
                                ),
                                params["branches"],
                            )
                        )

                        # NOTE: the branches are stored in the params
                        # in reverse order, so we need to reverse them
                        # This could totally be something which breaks in the future...
                        return jax.lax.cond(
                            Dual.tree_primal(in_vals[0]),
                            *reversed(branch_adev_functions),
                            key,
                            in_vals[1:],
                        )

                    # Default JVP rule for other JAX primitives.
                    else:
                        flat_primals, flat_tangents = ADInterpreter.flat_unzip(
                            Dual.tree_leaves(Dual.tree_pure(duals))
                        )
                        if len(flat_primals) == 0:
                            primal_outs = eqn.primitive.bind(*flat_primals, **params)
                            tangent_outs = jtu.tree_map(jnp.zeros_like, primal_outs)
                        else:
                            jvp = jax_autodiff.primitive_jvps.get(eqn.primitive)
                            if not jvp:
                                msg = f"differentiation rule for '{eqn.primitive}' not implemented"
                                raise NotImplementedError(msg)
                            primal_outs, tangent_outs = jvp(
                                flat_primals, flat_tangents, **params
                            )

                if not eqn.primitive.multiple_results:
                    primal_outs = [primal_outs]
                    tangent_outs = [tangent_outs]

                jax_util.safe_map(
                    dual_env.write,
                    eqn.outvars,
                    Dual.dual_tree(primal_outs, tangent_outs),
                )
            (out_dual,) = jax_util.safe_map(dual_env.read, jaxpr.outvars)
            if not isinstance(out_dual, Dual):
                out_dual = Dual(out_dual, jnp.zeros_like(out_dual))
            return out_dual

        return eval_jaxpr_iterate_dual(
            key, jaxpr.eqns, dual_env, jaxpr.invars, flat_duals
        )

    @staticmethod
    def forward_mode(f, kont=lambda v: v):
        def _inner(key, dual_tree: DualTree):
            primals = jtu.tree_leaves(Dual.tree_primal(dual_tree))
            closed_jaxpr, (_, _, out_tree) = stage(f)(*primals)
            jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.literals
            dual_leaves = Dual.tree_leaves(Dual.tree_pure(dual_tree))
            out_duals = ADInterpreter._eval_jaxpr_adev_jvp(
                key,
                jaxpr,
                consts,
                dual_leaves,
            )
            out_tree_def = out_tree()
            tree_primals, tree_tangents = Dual.tree_unzip(out_duals)
            out_dual_tree = Dual.dual_tree(
                jtu.tree_unflatten(out_tree_def, tree_primals),
                jtu.tree_unflatten(out_tree_def, tree_tangents),
            )
            vs = kont(out_dual_tree)
            return vs

        # Force coercion to JAX arrays.
        def maybe_array(v):
            return jnp.array(v, copy=False)

        def _dual(key, dual_tree: DualTree):
            dual_tree = jtu.tree_map(maybe_array, dual_tree)
            return _inner(key, dual_tree)

        return _dual


#################
# ADEV programs #
#################


@Pytree.dataclass
class ADEVProgram(Pytree):
    source: Callable[..., Any] = Pytree.static()

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,  # Pytree with Dual leaves.
        dual_kont: Callable[..., Any],
    ) -> Dual:
        def adev_jvp(f):
            @wraps(f)
            def wrapped(dual_tree: DualTree):
                return ADInterpreter.forward_mode(self.source, dual_kont)(
                    key, dual_tree
                )

            return wrapped

        return adev_jvp(self.source)(dual_tree)


###############
# Expectation #
###############


@Pytree.dataclass
class Expectation(Pytree):
    prog: ADEVProgram

    def jvp_estimate(self, key: PRNGKey, dual_tree: DualTree):
        # Trivial continuation.
        def _identity(v):
            return v

        return self.prog.jvp_estimate(key, dual_tree, _identity)

    def estimate(self, key, args):
        tangents = jtu.tree_map(lambda _: 0.0, args)
        return self.jvp_estimate(key, tangents).primal

    ##################################
    # JAX's native `grad` interface. #
    ##################################

    # The JVP rules here are registered below.
    # (c.f. Register custom forward mode with JAX)
    def grad_estimate(self, key: PRNGKey, primals: tuple[Any, ...]):
        def _invoke_closed_over(primals):
            return invoke_closed_over(self, key, primals)

        return jax.grad(_invoke_closed_over)(primals)


def expectation(source: Callable[..., Any]):
    prog = ADEVProgram(source)
    return Expectation(prog)


#########################################
# Register custom forward mode with JAX #
#########################################


# These two functions are defined to external to `Expectation`
# to ignore complexities with defining custom JVP rules for Pytree classes.
@jax.custom_jvp
def invoke_closed_over(instance, key, args):
    return instance.estimate(key, args)


def invoke_closed_over_jvp(primals, tangents):
    (instance, key, primals) = primals
    (_, _, tangents) = tangents
    duals = Dual.dual_tree(primals, tangents)
    out_dual = instance.jvp_estimate(key, duals)
    (v,), (tangent,) = Dual.tree_unzip(out_dual)
    return v, tangent


invoke_closed_over.defjvp(invoke_closed_over_jvp, symbolic_zeros=False)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from genjax._src.core.generative import (
    Argdiffs,
    Constraint,
    EditRequest,
    GenerativeFunction,
    Projection,
    Retdiff,
    Score,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.generative.choice_map import ChoiceMap
from genjax._src.core.interpreters.incremental import Diff, incremental
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    Generic,
    PRNGKey,
    TypeVar,
)

ArgTuple = TypeVar("ArgTuple", bound=tuple[Any, ...])
R = TypeVar("R")
S = TypeVar("S")


@Pytree.dataclass
class DimapTrace(Generic[R, S], Trace[S]):
    gen_fn: "DimapCombinator[Any, R, S]"
    inner: Trace[R]
    args: tuple[Any, ...]
    retval: S

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_gen_fn(self) -> GenerativeFunction[S]:
        return self.gen_fn

    def get_choices(self) -> ChoiceMap:
        return self.inner.get_choices()

    def get_retval(self) -> S:
        return self.retval

    def get_score(self) -> Score:
        return self.inner.get_score()


@Pytree.dataclass
class DimapCombinator(Generic[ArgTuple, R, S], GenerativeFunction[S]):
    """
    A combinator that transforms both the arguments and return values of a [`genjax.GenerativeFunction`][].

    This combinator allows for the modification of input arguments and return values through specified mapping functions, enabling the adaptation of a generative function to different contexts or requirements.

    Attributes:
        inner: The inner generative function to which the transformations are applied.
        argument_mapping: A function that maps the original arguments to the modified arguments that are passed to the inner generative function.
        retval_mapping: A function that takes a pair of `(args, return_value)` of the inner generative function and returns a mapped return value.
        info: Optional information or description about the specific instance of the combinator.

    Examples:
        Transforming the arguments and return values of a normal distribution draw via the [`genjax.dimap`][] decorator:
        ```python exec="yes" html="true" source="material-block" session="dimap"
        import genjax, jax


        @genjax.dimap(
            # double the mean and halve the std
            pre=lambda mean, std: (mean * 2, std / 2),
            post=lambda _args, retval: retval * 10,
        )
        @genjax.gen
        def transformed_normal_draw(mean, std):
            return genjax.normal(mean, std) @ "x"


        key = jax.random.key(314159)
        tr = jax.jit(transformed_normal_draw.simulate)(
            key,
            (
                0.0,  # Original mean
                1.0,  # Original std
            ),
        )
        print(tr.render_html())
        ```
    """

    inner: GenerativeFunction[R]
    argument_mapping: Callable[..., ArgTuple] = Pytree.static()
    retval_mapping: Callable[[tuple[Any, ...], ArgTuple, R], S] = Pytree.static()
    info: str | None = Pytree.static(default=None)

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> DimapTrace[R, S]:
        inner_args = self.argument_mapping(*args)
        tr = self.inner.simulate(key, inner_args)
        inner_retval = tr.get_retval()
        retval = self.retval_mapping(args, inner_args, inner_retval)
        return DimapTrace(self, tr, args, retval)

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[DimapTrace[R, S], Weight]:
        inner_args = self.argument_mapping(*args)
        tr, weight = self.inner.generate(key, constraint, inner_args)
        inner_retval = tr.get_retval()
        retval = self.retval_mapping(args, inner_args, inner_retval)
        return DimapTrace(self, tr, args, retval), weight

    def project(
        self,
        key: PRNGKey,
        trace: Trace[S],
        projection: Projection[Any],
    ) -> Weight:
        assert isinstance(trace, DimapTrace)
        return trace.inner.project(key, projection)

    def edit_change_target(
        self,
        key: PRNGKey,
        trace: Trace[S],
        request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[DimapTrace[R, S], Weight, Retdiff[S], EditRequest]:
        assert isinstance(trace, DimapTrace)

        primals = Diff.tree_primal(argdiffs)
        tangents = Diff.tree_tangent(argdiffs)

        inner_argdiffs = incremental(self.argument_mapping)(
            None,
            primals,
            tangents,
        )
        inner_trace: Trace[R] = trace.inner

        tr, w, inner_retdiff, bwd_request = self.inner.edit(
            key,
            inner_trace,
            request,
            inner_argdiffs,
        )

        inner_retval_primals = Diff.tree_primal(inner_retdiff)
        inner_retval_tangents = Diff.tree_tangent(inner_retdiff)

        def closed_mapping(args: tuple[Any, ...], retval: R) -> S:
            xformed_args = self.argument_mapping(*args)
            return self.retval_mapping(args, xformed_args, retval)

        retval_diff = incremental(closed_mapping)(
            None,
            (primals, inner_retval_primals),
            (tangents, inner_retval_tangents),
        )

        retval_primal: S = Diff.tree_primal(retval_diff)
        return (
            DimapTrace(self, tr, primals, retval_primal),
            w,
            retval_diff,
            bwd_request,
        )

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[S],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[DimapTrace[R, S], Weight, Retdiff[S], EditRequest]:
        assert isinstance(edit_request, Update)
        return self.edit_change_target(key, trace, edit_request, argdiffs)

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Score, S]:
        inner_args = self.argument_mapping(*args)
        w, inner_retval = self.inner.assess(sample, inner_args)
        retval = self.retval_mapping(args, inner_args, inner_retval)
        return w, retval


#############
# Decorator #
#############


def dimap(
    *,
    pre: Callable[..., ArgTuple] = lambda *args: args,
    post: Callable[[tuple[Any, ...], ArgTuple, R], S] = lambda _,
    _xformed,
    retval: retval,
    info: str | None = None,
) -> Callable[[GenerativeFunction[R]], DimapCombinator[ArgTuple, R, S]]:
    """
    Returns a decorator that wraps a [`genjax.GenerativeFunction`][] and applies pre- and post-processing functions to its arguments and return value.

    !!! info
        Prefer [`genjax.map`][] if you only need to transform the return value, or [`genjax.contramap`][] if you need to transform the arguments.

    Args:
        pre: A callable that preprocesses the arguments before passing them to the wrapped function. Note that `pre` must return a _tuple_ of arguments, not a bare argument. Default is the identity function.
        post: A callable that postprocesses the return value of the wrapped function. Default is the identity function.
        info: An optional string providing additional information about the `dimap` operation.

    Returns:
        A decorator that takes a [`genjax.GenerativeFunction`][] and returns a new [`genjax.GenerativeFunction`][] with the same behavior but with the arguments and return value transformed according to `pre` and `post`.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="dimap"
        import jax, genjax


        # Define pre- and post-processing functions
        def pre_process(x, y):
            return (x + 1, y * 2)


        def post_process(args, xformed, retval):
            return retval**2


        # Apply dimap to a generative function
        @genjax.dimap(pre=pre_process, post=post_process, info="Square of normal")
        @genjax.gen
        def dimap_model(x, y):
            return genjax.normal(x, y) @ "z"


        # Use the dimap model
        key = jax.random.key(0)
        trace = dimap_model.simulate(key, (2.0, 3.0))

        print(trace.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[R]) -> DimapCombinator[ArgTuple, R, S]:
        return DimapCombinator(f, pre, post, info)

    return decorator


def map(
    f: Callable[[R], S],
    *,
    info: str | None = None,
) -> Callable[[GenerativeFunction[R]], DimapCombinator[tuple[Any, ...], R, S]]:
    """
    Returns a decorator that wraps a [`genjax.GenerativeFunction`][] and applies a post-processing function to its return value.

    This is a specialized version of [`genjax.dimap`][] where only the post-processing function is applied.

    Args:
        f: A callable that postprocesses the return value of the wrapped function.
        info: An optional string providing additional information about the `map` operation.

    Returns:
        A decorator that takes a [`genjax.GenerativeFunction`][] and returns a new [`genjax.GenerativeFunction`][] with the same behavior but with the return value transformed according to `f`.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="map"
        import jax, genjax


        # Define a post-processing function
        def square(x):
            return x**2


        # Apply map to a generative function
        @genjax.map(square, info="Square of normal")
        @genjax.gen
        def map_model(x):
            return genjax.normal(x, 1.0) @ "z"


        # Use the map model
        key = jax.random.key(0)
        trace = map_model.simulate(key, (2.0,))

        print(trace.render_html())
        ```
    """

    def post(_args, _xformed, x: R) -> S:
        return f(x)

    return dimap(pre=lambda *args: args, post=post, info=info)


def contramap(
    f: Callable[..., ArgTuple],
    *,
    info: str | None = None,
) -> Callable[[GenerativeFunction[R]], DimapCombinator[ArgTuple, R, R]]:
    """
    Returns a decorator that wraps a [`genjax.GenerativeFunction`][] and applies a pre-processing function to its arguments.

    This is a specialized version of [`genjax.dimap`][] where only the pre-processing function is applied.

    Args:
        f: A callable that preprocesses the arguments of the wrapped function. Note that `f` must return a _tuple_ of arguments, not a bare argument.
        info: An optional string providing additional information about the `contramap` operation.

    Returns:
        A decorator that takes a [`genjax.GenerativeFunction`][] and returns a new [`genjax.GenerativeFunction`][] with the same behavior but with the arguments transformed according to `f`.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="contramap"
        import jax, genjax


        # Define a pre-processing function.
        # Note that this function must return a tuple of arguments!
        def add_one(x):
            return (x + 1,)


        # Apply contramap to a generative function
        @genjax.contramap(add_one, info="Add one to input")
        @genjax.gen
        def contramap_model(x):
            return genjax.normal(x, 1.0) @ "z"


        # Use the contramap model
        key = jax.random.key(0)
        trace = contramap_model.simulate(key, (2.0,))

        print(trace.render_html())
        ```
    """
    return dimap(pre=f, post=lambda _args, _xformed, ret: ret, info=info)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax.numpy as jnp
import jax.tree_util as jtu

from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    Constraint,
    EditRequest,
    GenerativeFunction,
    Mask,
    Projection,
    Retdiff,
    Score,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.interpreters.staging import FlagOp
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Flag,
    Generic,
    PRNGKey,
    ScalarFlag,
    TypeVar,
)

R = TypeVar("R")


@Pytree.dataclass
class MaskTrace(Generic[R], Trace[Mask[R]]):
    """A trace type for the `MaskCombinator` generative function.

    Users should use `MaskTrace.build` for constructing instances of `MaskTrace`,
    """

    mask_combinator: "MaskCombinator[R]"
    inner: Trace[R]
    args: tuple[Any, ...]
    chm: ChoiceMap
    score: Score
    ret: Mask[R]
    check: Flag

    @staticmethod
    def build(
        scan_gen_fn: "MaskCombinator[R]", inner: Trace[R], check: ScalarFlag
    ) -> "MaskTrace[R]":
        """Construct a new `MaskTrace` instance.

        This static method builds a `MaskTrace` by combining an inner trace with a check flag.
        The check flag determines whether the inner trace's choices and return value are masked.

        Args:
            scan_gen_fn: The MaskCombinator generative function
            inner: The inner trace to be masked
            check: A scalar boolean flag indicating whether to mask the inner trace

        Returns:
            A new MaskTrace instance with choices, return value and score masked with the check flag
        """
        # NOTE: constructing these values in `build`, where `check` is guaranteed scalar, allows us
        # to construct simple, non-vectorized `MaskTrace` instances. Returning these from `jax.vmap`
        # will allow JAX to construct a vectorized `MaskTrace` for us.
        #
        # If we instead deferred these computations to the methods (get_choices() etc), these methods would have to combine vectorized `self.inner.get_choices()` with a vectorized `self.check`. This is tricky and error-prone.
        args = (check, *inner.get_args())
        chm = inner.get_choices().mask(check)
        ret = Mask.build(inner.get_retval(), check)
        score = check * inner.get_score()

        return MaskTrace(scan_gen_fn, inner, args, chm, score, ret, check)

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_gen_fn(self):
        return self.mask_combinator

    def get_choices(self) -> ChoiceMap:
        return self.chm

    def get_retval(self):
        return self.ret

    def get_score(self):
        return self.score


@Pytree.dataclass
class MaskCombinator(Generic[R], GenerativeFunction[Mask[R]]):
    """
    Combinator which enables dynamic masking of generative functions. Takes a [`genjax.GenerativeFunction`][] and returns a new [`genjax.GenerativeFunction`][] which accepts an additional boolean first argument.

    If `True`, the invocation of the generative function is masked, and its contribution to the score is ignored. If `False`, it has the same semantics as if one was invoking the generative function without masking.

    The return value type is a `Mask`, with a flag value equal to the supplied boolean.

    Parameters:
        gen_fn: The generative function to be masked.

    Returns:
        The masked version of the input generative function.

    Examples:
        Masking a normal draw:
        ```python exec="yes" html="true" source="material-block" session="mask"
        import genjax, jax


        @genjax.mask
        @genjax.gen
        def masked_normal_draw(mean):
            return genjax.normal(mean, 1.0) @ "x"


        key = jax.random.key(314159)
        tr = jax.jit(masked_normal_draw.simulate)(
            key,
            (
                False,
                2.0,
            ),
        )
        print(tr.render_html())
        ```
    """

    gen_fn: GenerativeFunction[R]

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> MaskTrace[R]:
        check, inner_args = args[0], args[1:]
        tr = self.gen_fn.simulate(key, inner_args)
        return MaskTrace.build(self, tr, check)

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[MaskTrace[R], Weight]:
        check, inner_args = args[0], args[1:]

        tr, w = self.gen_fn.generate(key, constraint, inner_args)
        return MaskTrace.build(self, tr, check), w * check

    def project(
        self,
        key: PRNGKey,
        trace: Trace[Mask[R]],
        projection: Projection[Any],
    ) -> Weight:
        raise NotImplementedError

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[Mask[R]],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[MaskTrace[R], Weight, Retdiff[Mask[R]], EditRequest]:
        assert isinstance(trace, MaskTrace)
        assert isinstance(edit_request, Update)

        check_diff, inner_argdiffs = argdiffs[0], argdiffs[1:]
        post_check: ScalarFlag = Diff.tree_primal(check_diff)

        match trace:
            case MaskTrace():
                pre_check = trace.check
                original_trace: Trace[R] = trace.inner

        subrequest = Update(edit_request.constraint)

        premasked_trace, weight, retdiff, bwd_request = self.gen_fn.edit(
            key, original_trace, subrequest, inner_argdiffs
        )

        final_trace: Trace[R] = jtu.tree_map(
            lambda v1, v2: jnp.where(post_check, v1, v2),
            premasked_trace,
            original_trace,
        )

        t_to_t = FlagOp.and_(pre_check, post_check)
        t_to_f = FlagOp.and_(pre_check, FlagOp.not_(post_check))
        f_to_f = FlagOp.and_(FlagOp.not_(pre_check), FlagOp.not_(post_check))
        f_to_t = FlagOp.and_(FlagOp.not_(pre_check), post_check)

        final_weight = (
            #       What's the math for the weight term here?
            #
            # Well, if we started with a "masked false trace",
            # and then we flip the check_arg to True, we can re-use
            # the sampling process which created the original trace as
            # part of the move. The weight is the entire new trace's score.
            #
            # That's the transition False -> True:
            #
            #               final_weight = final_trace.score()
            #
            f_to_t * final_trace.get_score()
            #
            # On the other hand, if we started True, and went False, no matter
            # the update, we can make the choice that this move is just removing
            # the samples from the original trace, and ignoring the move.
            #
            # That's the transition True -> False:
            #
            #               final_weight = -original_trace.score()
            #
            + t_to_f * -original_trace.get_score()
            #
            # For the transition False -> False, we just ignore the move entirely.
            #
            #               final_weight = 0.0
            #
            + f_to_f * 0.0
            #
            # For the transition True -> True, we apply the move to the existing
            # unmasked trace. In that case, the weight is just the weight of the move.
            #
            #               final_weight = weight
            #
            + t_to_t * weight
            #
            # In any case, we always apply the move... we're not avoiding
            # that computation.
        )

        assert isinstance(bwd_request, Update)
        inner_chm = bwd_request.constraint

        return (
            MaskTrace.build(self, premasked_trace, post_check),
            final_weight,
            Mask.build(retdiff, check_diff),
            Update(
                inner_chm.mask(post_check),
            ),
        )

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Score, Mask[R]]:
        check, inner_args = args[0], args[1:]
        score, retval = self.gen_fn.assess(sample, inner_args)
        return (
            check * score,
            Mask(retval, check),
        )


#############
# Decorator #
#############


def mask(f: GenerativeFunction[R]) -> MaskCombinator[R]:
    """
    Combinator which enables dynamic masking of generative functions. Takes a [`genjax.GenerativeFunction`][] and returns a new [`genjax.GenerativeFunction`][] which accepts an additional boolean first argument.

    If `True`, the invocation of the generative function is masked, and its contribution to the score is ignored. If `False`, it has the same semantics as if one was invoking the generative function without masking.

    The return value type is a `Mask`, with a flag value equal to the supplied boolean.

    Args:
        f: The generative function to be masked.

    Returns:
        The masked version of the input generative function.

    Examples:
        Masking a normal draw:
        ```python exec="yes" html="true" source="material-block" session="mask"
        import genjax, jax


        @genjax.mask
        @genjax.gen
        def masked_normal_draw(mean):
            return genjax.normal(mean, 1.0) @ "x"


        key = jax.random.key(314159)
        tr = jax.jit(masked_normal_draw.simulate)(
            key,
            (
                False,
                2.0,
            ),
        )
        print(tr.render_html())
        ```
    """
    return MaskCombinator(f)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from genjax._src.generative_functions.distributions.custom.discrete_hmm import (
    DiscreteHMM,
    DiscreteHMMConfiguration,
    forward_filtering_backward_sampling,
)
from genjax._src.generative_functions.distributions.distribution import (
    Distribution,
    ExactDensity,
    ExactDensityFromCallables,
    exact_density,
)
from genjax._src.generative_functions.distributions.tensorflow_probability import (
    bates,
    bernoulli,
    beta,
    categorical,
    chi,
    chi2,
    dirichlet,
    exponential,
    flip,
    geometric,
    gumbel,
    half_cauchy,
    half_normal,
    half_student_t,
    inverse_gamma,
    kumaraswamy,
    laplace,
    logit_normal,
    moyal,
    multinomial,
    mv_normal,
    mv_normal_diag,
    negative_binomial,
    normal,
    plackett_luce,
    power_spherical,
    skellam,
    student_t,
    tfp_distribution,
    truncated_cauchy,
    truncated_normal,
    uniform,
    von_mises,
    von_mises_fisher,
    weibull,
    zipf,
)

__all__ = [
    "DiscreteHMM",
    "DiscreteHMMConfiguration",
    "Distribution",
    "ExactDensity",
    "ExactDensityFromCallables",
    "bates",
    "bernoulli",
    "beta",
    "categorical",
    "chi",
    "chi2",
    "dirichlet",
    "exact_density",
    "exponential",
    "flip",
    "forward_filtering_backward_sampling",
    "geometric",
    "gumbel",
    "half_cauchy",
    "half_normal",
    "half_student_t",
    "inverse_gamma",
    "kumaraswamy",
    "laplace",
    "logit_normal",
    "moyal",
    "multinomial",
    "mv_normal",
    "mv_normal_diag",
    "negative_binomial",
    "normal",
    "plackett_luce",
    "power_spherical",
    "skellam",
    "student_t",
    "tfp_distribution",
    "truncated_cauchy",
    "truncated_normal",
    "uniform",
    "von_mises",
    "von_mises_fisher",
    "weibull",
    "zipf",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A module containing a test suite for inference based on exact inference in hidden
Markov models (HMMs)."""

import jax
import jax.numpy as jnp

from genjax._src.core.generative import SelectionBuilder
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import FloatArray, IntArray, PRNGKey
from genjax._src.generative_functions.combinators.scan import (
    scan,
)
from genjax._src.generative_functions.distributions.custom.discrete_hmm import (
    DiscreteHMM,
    DiscreteHMMConfiguration,
)
from genjax._src.generative_functions.distributions.tensorflow_probability import (
    categorical,
)
from genjax._src.generative_functions.static import gen


@Pytree.dataclass
class DiscreteHMMInferenceProblem(Pytree):
    initial_state: IntArray
    log_posterior: FloatArray
    log_data_marginal: FloatArray
    latent_sequence: IntArray
    observation_sequence: IntArray


def build_test_against_exact_inference(
    max_length: int,
    state_space_size: IntArray,
    transition_distance_truncation: IntArray,
    observation_distance_truncation: IntArray,
    transition_variance: FloatArray,
    observation_variance: FloatArray,
):
    config = DiscreteHMMConfiguration(
        state_space_size,
        transition_distance_truncation,
        observation_distance_truncation,
        transition_variance,
        observation_variance,
    )

    @scan(n=max_length)
    @gen
    def markov_chain(state: IntArray, config: DiscreteHMMConfiguration):
        transition = config.transition_tensor()
        observation = config.observation_tensor()
        z = categorical(transition[state, :]) @ "z"
        _ = categorical(observation[z, :]) @ "x"
        return z, None

    def inference_test_generator(key: PRNGKey):
        key, sub_key = jax.random.split(key)
        initial_state = categorical.sample(sub_key, jnp.ones(config.linear_grid_dim))
        tr = markov_chain.simulate(sub_key, (max_length - 1, initial_state, config))
        z_sel = SelectionBuilder["z"]
        x_sel = SelectionBuilder["x"]
        latent_sequence = tr.get_choices().filter(z_sel)["z"]
        observation_sequence = tr.get_choices().filter(x_sel)["x"]
        log_data_marginal = DiscreteHMM.data_logpdf(config, observation_sequence)
        # This actually doesn't use any randomness.
        (log_posterior, _) = DiscreteHMM.estimate_logpdf(
            key, latent_sequence, config, observation_sequence
        )
        return DiscreteHMMInferenceProblem(
            initial_state,
            log_posterior,
            log_data_marginal,
            latent_sequence,
            observation_sequence,
        )

    return inference_test_generator



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains definitions for interpreters which act as program transformers
when staged out by JAX.

These interpreters support different patterns of program transformation. Mostly, each
implementation contains similar functionality, but are kept separate to allow
customization (here, I mean things like interpretation environments, or abstract types
which define the object-level values the interpreter produces for each primitive
statement in the Jaxpr).
"""



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from genjax._src.core.generative import GenerativeFunction
from genjax._src.core.typing import Any, ScalarFlag, TypeVar

R = TypeVar("R")


def or_else(
    if_gen_fn: GenerativeFunction[R],
    else_gen_fn: GenerativeFunction[R],
) -> GenerativeFunction[R]:
    """
    Given two [`genjax.GenerativeFunction`][]s `if_gen_fn` and `else_gen_fn`, returns a new [`genjax.GenerativeFunction`][] that accepts

    - a boolean argument
    - an argument tuple for `if_gen_fn`
    - an argument tuple for the supplied `else_gen_fn`

    and acts like `if_gen_fn` when the boolean is `True` or `else_gen_fn` otherwise.

    Args:
        else_gen_fn: called when the boolean argument is `False`.

    Returns:
        A [`genjax.GenerativeFunction`][] modified for conditional execution.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="or_else"
        import jax
        import jax.numpy as jnp
        import genjax


        @genjax.gen
        def if_model(x):
            return genjax.normal(x, 1.0) @ "if_value"


        @genjax.gen
        def else_model(x):
            return genjax.normal(x, 5.0) @ "else_value"


        or_else_model = genjax.or_else(if_model, else_model)


        @genjax.gen
        def model(toss: bool):
            # Note that `or_else_model` takes a new boolean predicate in
            # addition to argument tuples for each branch.
            return or_else_model(toss, (1.0,), (10.0,)) @ "tossed"


        key = jax.random.key(314159)

        tr = jax.jit(model.simulate)(key, (True,))

        print(tr.render_html())
        ```
    """

    def argument_mapping(
        b: ScalarFlag, if_args: tuple[Any, ...], else_args: tuple[Any, ...]
    ):
        # Note that `True` maps to 0 to select the "if" branch, `False` to 1.
        idx = jnp.array(jnp.logical_not(b), dtype=int)
        return (idx, if_args, else_args)

    return if_gen_fn.switch(else_gen_fn).contramap(
        argument_mapping, info="Derived combinator (OrElse)"
    )



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides an implementation of [ADEV: Sound Automatic Differentiation of Expected Values](https://dl.acm.org/doi/abs/10.1145/3571198), an AD algorithm which derivates forward mode derivative estimators for programs which denote expectations.

Our implementation tightly integrates with JAX's AD machinery, and extends the system described in the original paper with a reverse mode implementation, by utilizing JAX's support for automatically deriving a reverse mode from a forward mode (via [You Only Linearize Once: Tangents Transpose to Gradients](https://dl.acm.org/doi/abs/10.1145/3571236)).

Our implementation of the D{.} transformation from ADEV is structured around a separation of concerns of ADEV's transformation into two interpreters:

1. There's a partial CPS interpreter, which transforms a `Jaxpr` into a CPS-transformed `Jaxpr`, where certain primitives get access to their continuations.

2. There's a forward mode interpreter, which implements ADEV's forward mode AD for CPS-transformed `Jaxpr`s.
"""



# Copyright 2024 The MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typing
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from beartype.typing import overload
from jax import core as jc
from jax import tree_util as jtu
from jax.experimental import checkify
from jax.extend import linear_util as lu
from jax.interpreters import partial_eval as pe
from jax.util import safe_map

from genjax._src.checkify import optional_check
from genjax._src.core.typing import (
    Any,
    Array,
    ArrayLike,
    Callable,
    Flag,
    Iterable,
    Sequence,
    TypeVar,
    static_check_is_concrete,
)

if TYPE_CHECKING:
    import genjax

WrappedFunWithAux = tuple[lu.WrappedFun, Callable[[], Any]]

###############################
# Concrete Boolean arithmetic #
###############################

R = TypeVar("R")
F = TypeVar("F", bound=Callable[..., Any])


class FlagOp:
    """JAX compilation imposes restrictions on the control flow used in the compiled code.
    Branches gated by booleans must use GPU-compatible branching (e.g., `jax.lax.cond`).
    However, the GPU must compute both sides of the branch, wasting effort in the case
    where the gating boolean is constant. In such cases, if-based flow control will
    conceal the branch not taken from the JAX compiler, decreasing compilation time and
    code size for the result by not including the code for the branch that cannot be taken.

    This class centralizes the concrete short-cut logic used by GenJAX.
    """

    @staticmethod
    def is_scalar(f: Flag) -> bool:
        """Check if a flag is scalar.

        A flag is considered scalar if it is either a Python bool or a JAX array with empty shape ().

        Args:
            f: The flag to check. Can be a Python bool or JAX array.

        Returns:
            bool: True if the flag is scalar, False otherwise.
        """
        return isinstance(f, bool) or f.shape == ()

    @staticmethod
    @overload
    def and_(f: bool, g: bool) -> bool: ...

    @staticmethod
    @overload
    def and_(f: Array, g: bool | Array) -> Array: ...

    @staticmethod
    @overload
    def and_(f: bool | Array, g: Array) -> Array: ...

    @staticmethod
    def and_(f: Flag, g: Flag) -> Flag:
        if isinstance(f, bool) and isinstance(g, bool):
            return f & g
        else:
            return jnp.logical_and(f, g)

    @staticmethod
    @overload
    def or_(f: bool, g: bool) -> bool: ...

    @staticmethod
    @overload
    def or_(f: Array, g: bool | Array) -> Array: ...

    @staticmethod
    @overload
    def or_(f: bool | Array, g: Array) -> Array: ...

    @staticmethod
    def or_(f: Flag, g: Flag) -> Flag:
        if isinstance(f, bool) and isinstance(g, bool):
            return f | g
        else:
            return jnp.logical_or(f, g)

    @staticmethod
    @overload
    def xor_(f: bool, g: bool) -> bool: ...

    @staticmethod
    @overload
    def xor_(f: Array, g: bool | Array) -> Array: ...

    @staticmethod
    @overload
    def xor_(f: bool | Array, g: Array) -> Array: ...

    @staticmethod
    def xor_(f: Flag, g: Flag) -> Flag:
        if isinstance(f, bool) and isinstance(g, bool):
            return f ^ g
        else:
            return jnp.logical_xor(f, g)

    @staticmethod
    @overload
    def not_(f: bool) -> bool: ...

    @staticmethod
    @overload
    def not_(f: Array) -> Array: ...

    @staticmethod
    def not_(f: Flag) -> Flag:
        match f:
            case True:
                return False
            case False:
                return True
            case _:
                return jnp.logical_not(f)

    @staticmethod
    def concrete_true(f: Flag) -> bool:
        return f is True

    @staticmethod
    def concrete_false(f: Flag) -> bool:
        return f is False

    @staticmethod
    def where(f: Flag, tf: ArrayLike, ff: ArrayLike) -> ArrayLike:
        """Return tf or ff according to the truth value contained in flag
        in a manner that works in either the concrete or dynamic context"""
        if f is True:
            return tf
        if f is False:
            return ff
        return jax.lax.select(f, tf, ff)

    @staticmethod
    def cond(f: Flag, tf: Callable[..., R], ff: Callable[..., R], *args: Any) -> R:
        """Invokes `tf` with `args` if flag is true, else `ff`"""
        if f is True:
            return tf(*args)
        if f is False:
            return ff(*args)
        return jax.lax.cond(f, tf, ff, *args)


def staged_check(v):
    return static_check_is_concrete(v) and v


def tree_choose(
    idx: ArrayLike,
    pytrees: Sequence[R],
) -> R:
    """
    Version of `jax.numpy.choose` that

    - acts on lists of both `ArrayLike` and `Pytree` instances
    - acts like `vs[idx]` if `idx` is of type `int`.

    In the case of heterogenous types in `vs`, `tree_choose` will attempt to cast, or error if casting isn't possible. (mixed `bool` and `int` entries in `vs` will result in the cast of selected `bool` to `int`, for example.).

    Args:
        idx: The index used to select a value from `vs`.
        vs: A list of `Pytree` or `ArrayLike` values to choose from.

    Returns:
        The selected value from the list.
    """

    def inner(*vs: ArrayLike) -> ArrayLike:
        # Computing `result` above the branch allows us to:
        # - catch incompatible types / shapes in the result
        # - in the case of compatible types requiring casts (like bool => int),
        #   result's dtype tells us the final type.
        result = jnp.choose(idx, vs, mode="wrap")
        if isinstance(idx, int):
            return jnp.asarray(vs[idx % len(vs)], dtype=result.dtype)
        else:
            return result

    return jtu.tree_map(inner, *pytrees)


def multi_switch(
    idx, branches: Iterable[Callable[..., Any]], arg_tuples: Iterable[tuple[Any, ...]]
):
    """
    A wrapper around switch that allows selection between functions with differently-shaped return values.

    This function enables switching between branches that may have different output shapes.
    It creates a list of placeholder shapes for each branch and then uses a switch statement
    to select the appropriate function to fill in the correct shape.

    Args:
        idx: The index used to select the branch. If the index is out of bounds, it will be clamped to within bounds.
        branches: An iterable of callable functions representing different branches.
        arg_tuples: An iterable of argument tuples, one for each branch function.

    Returns:
        The result of calling the selected branch function with its corresponding arguments.

    Note:
        This function assumes that the number of branches matches the number of argument tuples.
        Each branch function should be able to handle its corresponding argument tuple.
    """

    def _make_setter(static_idx: int, f: Callable[..., Any], args: tuple[Any, ...]):
        def set_result(shapes: list[R]) -> list[R]:
            shapes[static_idx] = f(*args)
            return shapes

        return set_result

    pairs = list(zip(branches, arg_tuples))
    shapes = list(to_shape_fn(f, jnp.zeros)(*args) for f, args in pairs)
    fns = list(_make_setter(i, f, args) for i, (f, args) in enumerate(pairs))
    return jax.lax.switch(idx, fns, operand=shapes)


#########################
# Staged error handling #
#########################


def staged_err(check: Flag, msg, **kwargs):
    if FlagOp.concrete_true(check):
        raise Exception(msg)
    elif FlagOp.concrete_false(check):
        pass
    else:

        def _check():
            checkify.check(check, msg, **kwargs)

        optional_check(_check)


#######################################
# Staging utilities for type analysis #
#######################################


def get_shaped_aval(x):
    return jc.raise_to_shaped(jc.get_aval(x))


@lu.cache
def cached_stage_dynamic(flat_fun, in_avals):
    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals)
    typed_jaxpr = jc.ClosedJaxpr(jaxpr, consts)
    return typed_jaxpr


@lu.transformation_with_aux
def _flatten_fun_nokwargs(in_tree, *args_flat):
    py_args = jtu.tree_unflatten(in_tree, args_flat)
    ans = yield py_args, {}
    yield jtu.tree_flatten(ans)


# Wrapper to assign a correct type.
flatten_fun_nokwargs: Callable[[lu.WrappedFun, Any], WrappedFunWithAux] = (
    _flatten_fun_nokwargs  # pyright: ignore[reportAssignmentType]
)


def stage(f):
    """Returns a function that stages a function to a ClosedJaxpr."""

    def wrapped(*args, **kwargs):
        fun = lu.wrap_init(f, kwargs)
        flat_args, in_tree = jtu.tree_flatten(args)
        flat_fun, out_tree = flatten_fun_nokwargs(fun, in_tree)
        flat_avals = safe_map(get_shaped_aval, flat_args)
        typed_jaxpr = cached_stage_dynamic(flat_fun, tuple(flat_avals))
        return typed_jaxpr, (flat_args, in_tree, out_tree)

    return wrapped


def to_shape_fn(
    callable: F, fill_fn: Callable[[tuple[int], jnp.dtype[Any]], Array] | None = None
) -> F:
    """
    Convert a callable to a function that returns an empty pytree with the same structure as the original output (without any FLOPs).

    This function is similar to `jax.eval_shape`, but allows for optional post-processing of the output tree.

    Args:
        callable: The function to convert.
        fill_fn: A function to fill the output shapes with values. If None, returns the empty pytree as-is.
            The fill function takes a shape tuple and dtype as input and should return an array.

    Returns:
        A wrapped function that returns an empty pytree or a filled pytree with the same structure as the original function's output.
    """

    def wrapped(*args, **kwargs):
        shape = jax.eval_shape(callable, *args, **kwargs)
        if fill_fn is not None:
            f = fill_fn
            return jtu.tree_map(lambda x: f(x.shape, x.dtype), shape)
        else:
            return shape

    return typing.cast(F, wrapped)


_fake_key = jnp.array([0, 0], dtype=jnp.uint32)


def empty_trace(
    gen_fn: "genjax.GenerativeFunction[R]", args: "genjax.Arguments"
) -> "genjax.Trace[R]":
    """
    Create an empty trace for a generative function with given arguments (without spending any FLOPs).

    This function returns a trace with the same structure as a real trace, but filled with zero values. This is useful for static analysis and shape inference.

    Args:
        gen_fn: The generative function.
        args: The arguments to the generative function.

    Returns:
        A trace with the same structure as a real trace, but filled with zero values.
    """
    return to_shape_fn(gen_fn.simulate, jnp.zeros)(_fake_key, args)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    Constraint,
    EditRequest,
    GenerativeFunction,
    PrimitiveEditRequest,
    Projection,
    Regenerate,
    Retdiff,
    Score,
    Selection,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.generative.choice_map import ChoiceMapConstraint
from genjax._src.core.generative.functional_types import Mask
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    Flag,
    FloatArray,
    Generic,
    IntArray,
    PRNGKey,
    TypeVar,
)

Carry = TypeVar("Carry")
Y = TypeVar("Y")


@Pytree.dataclass
class ScanTrace(Generic[Carry, Y], Trace[tuple[Carry, Y]]):
    scan_gen_fn: "ScanCombinator[Carry, Y]"
    inner: Trace[tuple[Carry, Y]]
    args: tuple[Any, ...]
    retval: tuple[Carry, Y]
    score: FloatArray
    chm: ChoiceMap
    scan_length: int = Pytree.static()

    @staticmethod
    def build(
        scan_gen_fn: "ScanCombinator[Carry, Y]",
        inner: Trace[tuple[Carry, Y]],
        args: tuple[Any, ...],
        retval: tuple[Carry, Y],
        score: FloatArray,
        scan_length: int,
    ) -> "ScanTrace[Carry, Y]":
        if scan_length == 0:
            chm = ChoiceMap.empty()
        else:
            chm = jax.vmap(lambda tr: tr.get_choices())(inner)
        return ScanTrace(scan_gen_fn, inner, args, retval, score, chm, scan_length)

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_retval(self) -> tuple[Carry, Y]:
        return self.retval

    def get_choices(self) -> ChoiceMap:
        return self.chm

    def get_gen_fn(self):
        return self.scan_gen_fn

    def get_score(self):
        return self.score


@Pytree.dataclass(match_args=True)
class IndexRequest(PrimitiveEditRequest):
    """
    An `IndexRequest` is a primitive edit request which denotes a request to update a trace
    at a particular index of a vector combinator.

    The subrequest can be any type of `EditRequest`, the subrequest is responsible for enforcing or raising
    its own conditions for compositional usage.
    """

    idx: IntArray
    request: EditRequest


@Pytree.dataclass(match_args=True)
class VectorRequest(PrimitiveEditRequest):
    request: EditRequest


###################
# Scan combinator #
###################


@Pytree.dataclass
class ScanCombinator(Generic[Carry, Y], GenerativeFunction[tuple[Carry, Y]]):
    """`ScanCombinator` wraps a `kernel_gen_fn` [`genjax.GenerativeFunction`][]
    of type `(c, a) -> (c, b)` in a new [`genjax.GenerativeFunction`][] of type
    `(c, [a]) -> (c, [b])`, where.

    - `c` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - `a` may be a primitive, an array type or a pytree (container) type with array leaves
    - `b` may be a primitive, an array type or a pytree (container) type with array leaves.

    The values traced by each call to the original generative function will be nested under an integer index that matches the loop iteration index that generated it.

    For any array type specifier `t`, `[t]` represents the type with an additional leading axis, and if `t` is a pytree (container) type with array leaves then `[t]` represents the type with the same pytree structure and corresponding leaves each with an additional leading axis.

    When the type of `xs` in the snippet below (denoted `[a]` above) is an array type or None, and the type of `ys` in the snippet below (denoted `[b]` above) is an array type, the semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation:

    ```python
    def scan(f, init, xs, length=None):
        if xs is None:
            xs = [None] * length
        carry = init
        ys = []
        for x in xs:
            carry, y = f(carry, x)
            ys.append(y)
        return carry, np.stack(ys)
    ```

    Unlike that Python version, both `xs` and `ys` may be arbitrary pytree values, and so multiple arrays can be scanned over at once and produce multiple output arrays. `None` is actually a special case of this, as it represents an empty pytree.

    The loop-carried value `c` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `c` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Attributes:
        kernel_gen_fn: a generative function to be scanned of type `(c, a) -> (c, b)`, meaning that `f` accepts two arguments where the first is a value of the loop carry and the second is a slice of `xs` along its leading axis, and that `f` returns a pair where the first element represents a new value for the loop carry and the second represents a slice of the output.

        length: optional integer specifying the number of loop iterations, which (if supplied) must agree with the sizes of leading axes of the arrays in the returned function's second argument. If supplied then the returned generative function can take `None` as its second argument.

    Examples:
        Use the [`genjax.GenerativeFunction.scan`][] method:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax


        # A kernel_gen_fn generative function.
        @genjax.gen
        def random_walk_step(prev, _):
            x = genjax.normal(prev, 1.0) @ "x"
            return x, None


        init = 0.5
        key = jax.random.key(314159)

        random_walk = random_walk_step.scan(n=1000)

        tr = jax.jit(random_walk.simulate)(key, (init, None))
        print(tr.render_html())
        ```

        Or use the [`genjax.scan`][] decorator:
        ```python exec="yes" html="true" source="material-block" session="scan"
        @genjax.scan(n=1000)
        @genjax.gen
        def random_walk(prev, _):
            x = genjax.normal(prev, 1.0) @ "x"
            return x, None


        tr = jax.jit(random_walk.simulate)(key, (init, None))
        print(tr.render_html())
        ```
    """

    kernel_gen_fn: GenerativeFunction[tuple[Carry, Y]]

    # Only required for `None` carry inputs
    length: int | None = Pytree.static()

    # To get the type of return value, invoke the scanned over source (with abstract tracer arguments).
    def __abstract_call__(self, *args) -> tuple[Carry, Y]:
        return jax.lax.scan(
            self.kernel_gen_fn.__abstract_call__, *args, length=self.length
        )

    @staticmethod
    def _static_scan_length(xs: Any, length: int | None) -> int:
        # We start by triggering a scan to force all JAX validations to run.
        jax.lax.scan(lambda c, x: (c, None), None, xs, length=length)
        return length or jtu.tree_leaves(xs)[0].shape[0]

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> ScanTrace[Carry, Y]:
        carry, scanned_in = args

        def _inner(
            carry: tuple[PRNGKey, IntArray, Carry], scanned_in: Any
        ) -> tuple[
            tuple[PRNGKey, IntArray, Carry], tuple[Trace[tuple[Carry, Y]], Y, Score]
        ]:
            key, count, carried_value = carry
            key = jax.random.fold_in(key, count)

            tr = self.kernel_gen_fn.simulate(key, (carried_value, scanned_in))
            (carried_out, scanned_out) = tr.get_retval()
            score = tr.get_score()

            return (key, count + 1, carried_out), (tr, scanned_out, score)

        (_, _, carried_out), (tr, scanned_out, scores) = jax.lax.scan(
            _inner,
            (key, jnp.asarray(0), carry),
            scanned_in,
            length=self.length,
        )

        return ScanTrace.build(
            self,
            tr,
            args,
            (carried_out, scanned_out),
            jnp.sum(scores),
            self._static_scan_length(scanned_in, self.length),
        )

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[ScanTrace[Carry, Y], Weight]:
        assert isinstance(constraint, ChoiceMapConstraint)

        (carry, scanned_in) = args

        def _inner_generate(
            key: PRNGKey,
            constraint: Constraint,
            carry: Carry,
            scanned_in: Any,
        ) -> tuple[tuple[Carry, Score], tuple[Trace[tuple[Carry, Y]], Y, Weight]]:
            tr, w = self.kernel_gen_fn.generate(
                key,
                constraint,
                (carry, scanned_in),
            )
            (carry, scanned_out) = tr.get_retval()
            score = tr.get_score()
            return (carry, score), (tr, scanned_out, w)

        def _generate(
            carry: tuple[PRNGKey, IntArray, Carry],
            scanned_over: Any,
        ) -> tuple[
            tuple[PRNGKey, IntArray, Carry],
            tuple[Trace[tuple[Carry, Y]], Y, Score, Weight],
        ]:
            key, idx, carried_value = carry
            key = jax.random.fold_in(key, idx)
            submap = constraint.choice_map.get_submap(idx)
            subconstraint = ChoiceMapConstraint(submap)

            (carried_out, score), (tr, scanned_out, w) = _inner_generate(
                key, subconstraint, carried_value, scanned_over
            )

            return (key, idx + 1, carried_out), (tr, scanned_out, score, w)

        (_, _, carried_out), (tr, scanned_out, scores, ws) = jax.lax.scan(
            _generate,
            (key, jnp.asarray(0), carry),
            scanned_in,
            length=self.length,
        )
        return (
            ScanTrace[Carry, Y].build(
                self,
                tr,
                args,
                (carried_out, scanned_out),
                jnp.sum(scores),
                self._static_scan_length(scanned_in, self.length),
            ),
            jnp.sum(ws),
        )

    def project(
        self,
        key: PRNGKey,
        trace: Trace[tuple[Carry, Y]],
        projection: Projection[Any],
    ) -> Weight:
        assert isinstance(projection, Selection)
        assert isinstance(trace, ScanTrace)

        def _project(
            carry: tuple[PRNGKey, IntArray],
            subtrace: Trace[Any],
        ) -> tuple[tuple[PRNGKey, IntArray], Weight]:
            key, idx = carry
            key = jax.random.fold_in(key, idx)
            subprojection = projection(idx)
            assert isinstance(subprojection, Selection)
            w = subtrace.project(
                key,
                subprojection,
            )

            return (key, idx + 1), w

        (_, _), ws = jax.lax.scan(
            _project,
            (key, jnp.asarray(0)),
            trace.inner,
            length=self.length,
        )
        return jnp.sum(ws)

    def edit_index(
        self,
        key: PRNGKey,
        trace: ScanTrace[Carry, Y],
        idx: IntArray,
        request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[ScanTrace[Carry, Y], Weight, Retdiff[tuple[Carry, Y]], EditRequest]:
        # For now, we don't allow changes to the arguments for this type of edit.
        assert Diff.static_check_no_change(argdiffs)

        (_, scanned_argdiff) = argdiffs
        scanned_in = Diff.tree_primal(scanned_argdiff)
        (old_carried_out, old_scanned_out) = trace.get_retval()
        trace_slice: Trace[tuple[Carry, Y]] = jtu.tree_map(
            lambda v: v[idx], trace.inner
        )
        new_slice_trace, w, retdiff, bwd_request = request.edit(
            key, trace_slice, Diff.no_change(trace_slice.get_args())
        )
        (carry_retdiff, scanned_retdiff) = retdiff
        next_slice, next_scanned_in = jtu.tree_map(
            lambda v: v[idx + 1], (trace.inner, scanned_in)
        )

        # The Update edit request here is used to force a visitation of
        # generative function callees which could be affected by the changes
        # above -- to account for new scores, and changes reflected in the weight.
        next_request = Update(ChoiceMap.empty())
        next_slice_trace, next_w, retdiff, _ = next_request.edit(
            key,
            next_slice,
            (carry_retdiff, Diff.no_change(next_scanned_in)),
        )

        # The carry value better not have changed, otherwise
        # the assumptions of this edit are not satisfied.
        #
        # Here, we could allow the carry from the slice to flow out via the scanned out
        # value in the next slice, but we just disallow that for now for simplicity.
        assert Diff.static_check_no_change(retdiff)

        idx_array = jnp.arange(trace.scan_length)
        slice_scanned_out = Diff.tree_primal(scanned_retdiff)
        new_scanned_out: Y = jtu.tree_map(
            lambda v1, v2: jnp.where(idx_array == idx, v1, v2),
            slice_scanned_out,
            old_scanned_out,
        )
        new_scanned_retdiff = Diff.unknown_change(new_scanned_out)
        max_length = self._static_scan_length(scanned_in, self.length)

        # Mutate the trace in-place.
        # Underneath JIT, should compile to an in-place update.
        def mutator(v, idx, setter):
            return v.at[idx].set(
                jnp.where(idx < max_length, setter, v[idx]),
            )

        new_inner_trace = jtu.tree_map(
            lambda v, v_: mutator(v, idx, v_), trace.inner, new_slice_trace
        )
        new_inner_trace = jtu.tree_map(
            lambda v, v_: mutator(v, idx + 1, v_), new_inner_trace, next_slice_trace
        )
        scores = jax.vmap(lambda tr: tr.get_score())(new_inner_trace)

        # We don't actually know if the index which was updated was the last one.
        # Therefore, we need to provide a where selection
        # between the carry from index, and the next slice --
        carry_out = Diff.tree_primal(carry_retdiff)
        carry_out_ = Diff.tree_primal(retdiff[0])
        carried_out = jtu.tree_map(
            lambda v, v_: jnp.where(idx < max_length, v_, v),
            carry_out,
            carry_out_,
        )

        return (
            ScanTrace[Carry, Y].build(
                self,
                new_inner_trace,
                Diff.tree_primal(argdiffs),
                (carried_out, new_scanned_out),
                jnp.sum(scores),
                self._static_scan_length(scanned_in, self.length),
            ),
            w + (next_w * (idx + 1 < max_length)),
            # We always set the carried out value to be an unknown change, conservatively.
            (Diff.unknown_change(old_carried_out), new_scanned_retdiff),
            IndexRequest(idx, bwd_request),
        )

    def edit_regenerate(
        self,
        key: PRNGKey,
        trace: ScanTrace[Carry, Y],
        selection: Selection,
        argdiffs: Argdiffs,
    ) -> tuple[ScanTrace[Carry, Y], Weight, Retdiff[tuple[Carry, Y]], EditRequest]:
        diffs = Diff.unknown_change(Diff.tree_primal(argdiffs))
        carry_diff: Carry = diffs[0]
        scanned_in_diff: Any = diffs[1:]

        def _inner_edit(
            key: PRNGKey,
            subtrace: Trace[tuple[Carry, Y]],
            subselection: Selection,
            carry: Carry,
            scanned_in: Any,
        ) -> tuple[
            tuple[Carry, Score],
            tuple[Trace[tuple[Carry, Y]], Retdiff[Y], Weight, EditRequest],
        ]:
            request = Regenerate(subselection)
            (
                new_subtrace,
                w,
                kernel_retdiff,
                bwd_request,
            ) = request.edit(
                key,
                subtrace,
                (carry, scanned_in),
            )
            (carry_retdiff, scanned_out_retdiff) = Diff.unknown_change(kernel_retdiff)
            score = new_subtrace.get_score()
            return (carry_retdiff, score), (
                new_subtrace,
                scanned_out_retdiff,
                w,
                bwd_request,
            )

        def _edit(
            carry: tuple[PRNGKey, IntArray, Carry],
            scanned_over: tuple[Trace[tuple[Carry, Y]], Any],
        ) -> tuple[
            tuple[PRNGKey, IntArray, Carry],
            tuple[Trace[tuple[Carry, Y]], Retdiff[Y], Score, Weight, EditRequest],
        ]:
            key, idx, carried_value = carry
            subtrace, scanned_in = scanned_over
            key = jax.random.fold_in(key, idx)
            subselection = selection(idx)
            (
                (carried_out, score),
                (new_subtrace, scanned_out, w, inner_bwd_request),
            ) = _inner_edit(key, subtrace, subselection, carried_value, scanned_in)

            return (key, idx + 1, carried_out), (
                new_subtrace,
                scanned_out,
                score,
                w,
                inner_bwd_request,
            )

        (
            (_, _, carried_out_diff),
            (new_subtraces, scanned_out_diff, scores, ws, bwd_constraints),
        ) = jax.lax.scan(
            _edit,
            (key, jnp.asarray(0), carry_diff),
            (trace.inner, *scanned_in_diff),
            length=self.length,
        )
        carried_out, scanned_out = Diff.tree_primal((
            carried_out_diff,
            scanned_out_diff,
        ))
        return (
            ScanTrace.build(
                self,
                new_subtraces,
                Diff.tree_primal(argdiffs),
                (carried_out, scanned_out),
                jnp.sum(scores),
                trace.scan_length,
            ),
            jnp.sum(ws),
            (carried_out_diff, scanned_out_diff),
            VectorRequest(bwd_constraints),
        )

    def edit_update(
        self,
        key: PRNGKey,
        trace: ScanTrace[Carry, Y],
        constraint: ChoiceMap,
        argdiffs: Argdiffs,
    ) -> tuple[ScanTrace[Carry, Y], Weight, Retdiff[tuple[Carry, Y]], EditRequest]:
        diffs = Diff.unknown_change(Diff.tree_primal(argdiffs))
        carry_diff: Carry = diffs[0]
        scanned_in_diff: Any = diffs[1:]

        def _inner_edit(
            key: PRNGKey,
            subtrace: Trace[tuple[Carry, Y]],
            subconstraint: ChoiceMap,
            carry: Carry,
            scanned_in: Any,
        ) -> tuple[
            tuple[Carry, Score],
            tuple[Trace[tuple[Carry, Y]], Retdiff[Y], Weight, EditRequest],
        ]:
            (
                new_subtrace,
                w,
                kernel_retdiff,
                bwd_request,
            ) = self.kernel_gen_fn.edit(
                key,
                subtrace,
                Update(subconstraint),
                (carry, scanned_in),
            )
            (carry_retdiff, scanned_out_retdiff) = Diff.unknown_change(kernel_retdiff)
            score = new_subtrace.get_score()
            return (carry_retdiff, score), (
                new_subtrace,
                scanned_out_retdiff,
                w,
                bwd_request,
            )

        def _edit(
            carry: tuple[PRNGKey, IntArray, Carry],
            scanned_over: tuple[Trace[tuple[Carry, Y]], Any],
        ) -> tuple[
            tuple[PRNGKey, IntArray, Carry],
            tuple[Trace[tuple[Carry, Y]], Retdiff[Y], Score, Weight, ChoiceMap],
        ]:
            key, idx, carried_value = carry
            subtrace, scanned_in = scanned_over
            key = jax.random.fold_in(key, idx)
            subconstraint = constraint(idx)
            assert isinstance(subconstraint, ChoiceMap)
            (
                (carried_out, score),
                (new_subtrace, scanned_out, w, inner_bwd_request),
            ) = _inner_edit(key, subtrace, subconstraint, carried_value, scanned_in)
            assert isinstance(inner_bwd_request, Update)
            bwd_chm = inner_bwd_request.constraint

            return (key, idx + 1, carried_out), (
                new_subtrace,
                scanned_out,
                score,
                w,
                bwd_chm,
            )

        (
            (_, _, carried_out_diff),
            (new_subtraces, scanned_out_diff, scores, ws, bwd_constraints),
        ) = jax.lax.scan(
            _edit,
            (key, jnp.asarray(0), carry_diff),
            (trace.inner, *scanned_in_diff),
            length=self.length,
        )
        carried_out, scanned_out = Diff.tree_primal((
            carried_out_diff,
            scanned_out_diff,
        ))
        return (
            ScanTrace.build(
                self,
                new_subtraces,
                Diff.tree_primal(argdiffs),
                (carried_out, scanned_out),
                jnp.sum(scores),
                trace.scan_length,
            ),
            jnp.sum(ws),
            (carried_out_diff, scanned_out_diff),
            Update(bwd_constraints),
        )

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[tuple[Carry, Y]],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[ScanTrace[Carry, Y], Weight, Retdiff[tuple[Carry, Y]], EditRequest]:
        assert isinstance(trace, ScanTrace)
        match edit_request:
            case Regenerate(selection):
                return self.edit_regenerate(
                    key,
                    trace,
                    selection,
                    argdiffs,
                )
            case Update(constraint):
                return self.edit_update(
                    key,
                    trace,
                    constraint,
                    argdiffs,
                )
            case IndexRequest(idx, subrequest):
                return self.edit_index(
                    key,
                    trace,
                    idx,
                    subrequest,
                    argdiffs,
                )
            case _:
                raise NotImplementedError

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Score, Any]:
        (carry, scanned_in) = args

        def _assess(carry, scanned_in):
            idx, carried_value = carry
            submap = sample.get_submap(idx)

            score, (carry, scanned_out) = self.kernel_gen_fn.assess(
                submap, (carried_value, scanned_in)
            )

            return (idx + 1, carry), (scanned_out, score)

        (_, carried_out), (scanned_out, scores) = jax.lax.scan(
            _assess,
            (0, carry),
            scanned_in,
            length=self.length,
        )
        return (
            jnp.sum(scores),
            (carried_out, scanned_out),
        )


##############
# Decorators #
##############


def scan(
    *, n: int | None = None
) -> Callable[
    [GenerativeFunction[tuple[Carry, Y]]], GenerativeFunction[tuple[Carry, Y]]
]:
    """Returns a decorator that wraps a [`genjax.GenerativeFunction`][] of type
    `(c, a) -> (c, b)`and returns a new [`genjax.GenerativeFunction`][] of type
    `(c, [a]) -> (c, [b])` where.

    - `c` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - `a` may be a primitive, an array type or a pytree (container) type with array leaves
    - `b` may be a primitive, an array type or a pytree (container) type with array leaves.

    The values traced by each call to the original generative function will be nested under an integer index that matches the loop iteration index that generated it.

    For any array type specifier `t`, `[t]` represents the type with an additional leading axis, and if `t` is a pytree (container) type with array leaves then `[t]` represents the type with the same pytree structure and corresponding leaves each with an additional leading axis.

    When the type of `xs` in the snippet below (denoted `[a]` above) is an array type or None, and the type of `ys` in the snippet below (denoted `[b]` above) is an array type, the semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation:

    ```python
    def scan(f, init, xs, length=None):
        if xs is None:
            xs = [None] * length
        carry = init
        ys = []
        for x in xs:
            carry, y = f(carry, x)
            ys.append(y)
        return carry, np.stack(ys)
    ```

    Unlike that Python version, both `xs` and `ys` may be arbitrary pytree values, and so multiple arrays can be scanned over at once and produce multiple output arrays. `None` is actually a special case of this, as it represents an empty pytree.

    The loop-carried value `c` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `c` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Args:
        n: optional integer specifying the number of loop iterations, which (if supplied) must agree with the sizes of leading axes of the arrays in the returned function's second argument. If supplied then the returned generative function can take `None` as its second argument.

    Returns:
        A new [`genjax.GenerativeFunction`][] that takes a loop-carried value and a new input, and returns a new loop-carried value along with either `None` or an output to be collected into the second return value.

    Examples:
        Scan for 1000 iterations with no array input:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax


        @genjax.scan(n=1000)
        @genjax.gen
        def random_walk(prev, _):
            x = genjax.normal(prev, 1.0) @ "x"
            return x, None


        init = 0.5
        key = jax.random.key(314159)

        tr = jax.jit(random_walk.simulate)(key, (init, None))
        print(tr.render_html())
        ```

        Scan across an input array:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax.numpy as jnp


        @genjax.scan()
        @genjax.gen
        def add_and_square_all(sum, x):
            new_sum = sum + x
            return new_sum, sum * sum


        init = 0.0
        xs = jnp.ones(10)

        tr = jax.jit(add_and_square_all.simulate)(key, (init, xs))

        # The retval has the final carry and an array of all `sum*sum` returned.
        print(tr.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[tuple[Carry, Y]]):
        return ScanCombinator[Carry, Y](f, length=n)

    return decorator


def prepend_initial_acc(
    args: tuple[Carry, ...], _: tuple[Carry, ...], ret: tuple[Carry, Carry]
) -> Carry:
    """Prepends the initial accumulator value to the array of accumulated
    values.

    This function is used in the context of scan operations to include the initial
    accumulator state in the output, effectively providing a complete history of
    the accumulator's values throughout the scan.

    Args:
        args: A tuple containing the initial arguments to the scan operation. The first element is expected to be the initial accumulator value.
        ret: A tuple containing the final accumulator value and an array of intermediate accumulator values from the scan operation.

    Returns:
        A tree structure where each leaf is an array with the initial accumulator value prepended to the corresponding array of intermediate values.

    Note:
        This function uses JAX's tree mapping to handle nested structures in the accumulator, allowing it to work with complex accumulator types.
    """
    init_acc = args[0]
    xs = ret[1]

    def cat(init, arr):
        return jnp.concatenate([jnp.array(init)[jnp.newaxis], arr])

    return jax.tree.map(cat, init_acc, xs)


def accumulate() -> Callable[[GenerativeFunction[Carry]], GenerativeFunction[Carry]]:
    """Returns a decorator that wraps a [`genjax.GenerativeFunction`][] of type
    `(c, a) -> c` and returns a new [`genjax.GenerativeFunction`][] of type
    `(c, [a]) -> [c]` where.

    - `c` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - `[c]` is an array of all loop-carried values seen during iteration (including the first)
    - `a` may be a primitive, an array type or a pytree (container) type with array leaves

    All traced values are nested under an index.

    For any array type specifier `t`, `[t]` represents the type with an additional leading axis, and if `t` is a pytree (container) type with array leaves then `[t]` represents the type with the same pytree structure and corresponding leaves each with an additional leading axis.

    The semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation (note the similarity to [`itertools.accumulate`](https://docs.python.org/3/library/itertools.html#itertools.accumulate)):

    ```python
    def accumulate(f, init, xs):
        carry = init
        carries = [init]
        for x in xs:
            carry = f(carry, x)
            carries.append(carry)
        return carries
    ```

    Unlike that Python version, both `xs` and `carries` may be arbitrary pytree values, and so multiple arrays can be scanned over at once and produce multiple output arrays.

    The loop-carried value `c` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `c` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Examples:
        accumulate a running total:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax
        import jax.numpy as jnp


        @genjax.accumulate()
        @genjax.gen
        def add(sum, x):
            new_sum = sum + x
            return new_sum


        init = 0.0
        key = jax.random.key(314159)
        xs = jnp.ones(10)

        tr = jax.jit(add.simulate)(key, (init, xs))
        print(tr.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[Carry]) -> GenerativeFunction[Carry]:
        return (
            f.map(lambda ret: (ret, ret))
            .scan()
            .dimap(pre=lambda *args: args, post=prepend_initial_acc)
        )

    return decorator


def reduce() -> Callable[[GenerativeFunction[Carry]], GenerativeFunction[Carry]]:
    """Returns a decorator that wraps a [`genjax.GenerativeFunction`][] of type
    `(c, a) -> c` and returns a new [`genjax.GenerativeFunction`][] of type
    `(c, [a]) -> c` where.

    - `c` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - `a` may be a primitive, an array type or a pytree (container) type with array leaves

    All traced values are nested under an index.

    For any array type specifier `t`, `[t]` represents the type with an additional leading axis, and if `t` is a pytree (container) type with array leaves then `[t]` represents the type with the same pytree structure and corresponding leaves each with an additional leading axis.

    The semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation (note the similarity to [`functools.reduce`](https://docs.python.org/3/library/itertools.html#functools.reduce)):

    ```python
    def reduce(f, init, xs):
        carry = init
        for x in xs:
            carry = f(carry, x)
        return carry
    ```

    Unlike that Python version, both `xs` and `carry` may be arbitrary pytree values, and so multiple arrays can be scanned over at once and produce multiple output arrays.

    The loop-carried value `c` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `c` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Examples:
        sum an array of numbers:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax
        import jax.numpy as jnp


        @genjax.reduce()
        @genjax.gen
        def add(sum, x):
            new_sum = sum + x
            return new_sum


        init = 0.0
        key = jax.random.key(314159)
        xs = jnp.ones(10)

        tr = jax.jit(add.simulate)(key, (init, xs))
        print(tr.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[Carry]) -> GenerativeFunction[Carry]:
        def pre(ret: Carry):
            return ret, None

        def post(ret: tuple[Carry, None]):
            return ret[0]

        return f.map(pre).scan().map(post)

    return decorator


def iterate(*, n: int) -> Callable[[GenerativeFunction[Y]], GenerativeFunction[Y]]:
    """Returns a decorator that wraps a [`genjax.GenerativeFunction`][] of type
    `a -> a` and returns a new [`genjax.GenerativeFunction`][] of type `a ->
    [a]` where.

    - `a` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - `[a]` is an array of all `a`, `f(a)`, `f(f(a))` etc. values seen during iteration.

    All traced values are nested under an index.

    The semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation:

    ```python
    def iterate(f, n, init):
        input = init
        seen = [init]
        for _ in range(n):
            input = f(input)
            seen.append(input)
        return seen
    ```

    `init` may be an arbitrary pytree value, and so multiple arrays can be iterated over at once and produce multiple output arrays.

    The iterated value `a` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `a` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Args:
        n: the number of iterations to run.

    Examples:
        iterative addition, returning all intermediate sums:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax


        @genjax.iterate(n=100)
        @genjax.gen
        def inc(x):
            return x + 1


        init = 0.0
        key = jax.random.key(314159)

        tr = jax.jit(inc.simulate)(key, (init,))
        print(tr.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[Y]) -> GenerativeFunction[Y]:
        # strip off the JAX-supplied `None` on the way in, accumulate `ret` on the way out.
        return (
            f.dimap(
                pre=lambda *args: args[:-1],
                post=lambda _args, _xformed, ret: (ret, ret),
            )
            .scan(n=n)
            .dimap(pre=lambda *args: (*args, None), post=prepend_initial_acc)
        )

    return decorator


def iterate_final(
    *, n: int
) -> Callable[[GenerativeFunction[Y]], GenerativeFunction[Y]]:
    """Returns a decorator that wraps a [`genjax.GenerativeFunction`][] of type
    `a -> a` and returns a new [`genjax.GenerativeFunction`][] of type `a -> a`
    where.

    - `a` is a loop-carried value, which must hold a fixed shape and dtype across all iterations
    - the original function is invoked `n` times with each input coming from the previous invocation's output, so that the new function returns $f^n(a)$

    All traced values are nested under an index.

    The semantics of the returned [`genjax.GenerativeFunction`][] are given roughly by this Python implementation:

    ```python
    def iterate_final(f, n, init):
        ret = init
        for _ in range(n):
            ret = f(ret)
        return ret
    ```

    `init` may be an arbitrary pytree value, and so multiple arrays can be iterated over at once and produce multiple output arrays.

    The iterated value `a` must hold a fixed shape and dtype across all iterations (and not just be consistent up to NumPy rank/shape broadcasting and dtype promotion rules, for example). In other words, the type `a` in the type signature above represents an array with a fixed shape and dtype (or a nested tuple/list/dict container data structure with a fixed structure and arrays with fixed shape and dtype at the leaves).

    Args:
        n: the number of iterations to run.

    Examples:
        iterative addition:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax


        @genjax.iterate_final(n=100)
        @genjax.gen
        def inc(x):
            return x + 1


        init = 0.0
        key = jax.random.key(314159)

        tr = jax.jit(inc.simulate)(key, (init,))
        print(tr.render_html())
        ```
    """

    def decorator(f: GenerativeFunction[Y]) -> GenerativeFunction[Y]:
        # strip off the JAX-supplied `None` on the way in, no accumulation on the way out.
        def pre_post(_, _xformed, ret: Y):
            return ret, None

        def post_post(_, _xformed, ret: tuple[Y, None]):
            return ret[0]

        return (
            f.dimap(pre=lambda *args: args[:-1], post=pre_post)
            .scan(n=n)
            .dimap(pre=lambda *args: (*args, None), post=post_post)
        )

    return decorator


# Masked Combinators


def masked_iterate_final() -> Callable[[GenerativeFunction[Y]], GenerativeFunction[Y]]:
    """
    Transforms a generative function that takes a single argument of type `a` and returns a value of type `a`, into a function that takes a tuple of arguments `(a, [mask])` and returns a value of type `a`.

    The original function is modified to accept an additional argument `mask`, which is a boolean value indicating whether the operation should be masked or not. The function returns the result of the original operation if `mask` is `True`, and the original input if `mask` is `False`.

    All traced values from the kernel generative function are traced (with an added axis due to the scan) but only those indices from [mask] with a flag of True will accounted for in inference, notably for score computations.

    Example:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax

        masks = jnp.array([True, False, True])


        # Create a kernel generative function
        @genjax.gen
        def step(x):
            _ = (
                genjax.normal.mask().vmap(in_axes=(0, None, None))(masks, x, 1.0)
                @ "rats"
            )
            return x


        # Create a model using masked_iterate_final
        model = genjax.masked_iterate_final()(step)

        # Simulate from the model
        key = jax.random.key(0)
        mask_steps = jnp.arange(10) < 5
        tr = model.simulate(key, (0.0, mask_steps))
        print(tr.render_html())
        ```
    """

    def decorator(step: GenerativeFunction[Y]) -> GenerativeFunction[Y]:
        def pre(state, flag: Flag):
            return flag, state

        def post(_unused_args, _xformed, masked_retval: Mask[Y]):
            return masked_retval.value, None

        # scan_step: (a, bool) -> a
        scan_step = step.mask().dimap(pre=pre, post=post)
        return scan_step.scan().map(lambda ret: ret[0])

    return decorator


def masked_iterate() -> Callable[[GenerativeFunction[Y]], GenerativeFunction[Y]]:
    """
    Transforms a generative function that takes a single argument of type `a` and returns a value of type `a`, into a function that takes a tuple of arguments `(a, [mask])` and returns a list of values of type `a`.

    The original function is modified to accept an additional argument `mask`, which is a boolean value indicating whether the operation should be masked or not. The function returns a Masked list of results of the original operation with the input [mask] as mask.

    All traced values from the kernel generative function are traced (with an added axis due to the scan) but only those indices from [mask] with a flag of True will accounted for in inference, notably for score computations.

    Example:
        ```python exec="yes" html="true" source="material-block" session="scan"
        import jax
        import genjax

        masks = jnp.array([True, False, True])


        # Create a kernel generative function
        @genjax.gen
        def step(x):
            _ = (
                genjax.normal.mask().vmap(in_axes=(0, None, None))(masks, x, 1.0)
                @ "rats"
            )
            return x


        # Create a model using masked_iterate
        model = genjax.masked_iterate()(step)

        # Simulate from the model
        key = jax.random.key(0)
        mask_steps = jnp.arange(10) < 5
        tr = model.simulate(key, (0.0, mask_steps))
        print(tr.render_html())
        ```
    """

    def decorator(step: GenerativeFunction[Y]) -> GenerativeFunction[Y]:
        def pre(state, flag: Flag):
            return flag, state

        def post(_unused_args, _xformed, masked_retval: Mask[Y]):
            v = masked_retval.value
            return v, v

        # scan_step: (a, bool) -> a
        scan_step = step.mask().dimap(pre=pre, post=post)
        return scan_step.scan().dimap(pre=lambda *args: args, post=prepend_initial_acc)

    return decorator



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from genjax._src.core.generative import GenerativeFunction
from genjax._src.core.typing import TypeVar
from genjax._src.generative_functions.combinators.switch import (
    switch,
)
from genjax._src.generative_functions.distributions.tensorflow_probability import (
    categorical,
)
from genjax._src.generative_functions.static import gen

R = TypeVar("R")


def mix(*gen_fns: GenerativeFunction[R]) -> GenerativeFunction[R]:
    """
    Creates a mixture model from a set of generative functions.

    This function takes multiple generative functions as input and returns a new generative function that represents a mixture model.

    The returned generative function takes the following arguments:

    - `mixture_logits`: Logits for the categorical distribution used to select a component.
    - `*args`: Argument tuples for each of the input generative functions

    and samples from one of the input generative functions based on draw from a categorical distribution defined by the provided mixture logits.

    Args:
        *gen_fns: Variable number of [`genjax.GenerativeFunction`][]s to be mixed.

    Returns:
        A new [`genjax.GenerativeFunction`][] representing the mixture model.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="mix"
        import jax
        import genjax


        # Define component generative functions
        @genjax.gen
        def component1(x):
            return genjax.normal(x, 1.0) @ "y"


        @genjax.gen
        def component2(x):
            return genjax.normal(x, 2.0) @ "y"


        # Create mixture model
        mixture = genjax.mix(component1, component2)

        # Use the mixture model
        key = jax.random.key(0)
        logits = jax.numpy.array([0.3, 0.7])  # Favors component2
        trace = mixture.simulate(key, (logits, (0.0,), (7.0,)))
        print(trace.render_html())
        ```
    """

    inner_combinator_closure = switch(*gen_fns)

    @gen
    def mixture_model(mixture_logits, *args) -> R:
        mix_idx = categorical(mixture_logits) @ "mixture_component"
        v = inner_combinator_closure(mix_idx, *args) @ "component_sample"
        return v

    return mixture_model



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax.numpy as jnp
import jax.random as jrand

from genjax._src.core.generative.choice_map import (
    ChoiceMap,
    Selection,
)
from genjax._src.core.generative.core import (
    Argdiffs,
    EditRequest,
    PrimitiveEditRequest,
    Retdiff,
    Weight,
)
from genjax._src.core.generative.generative_function import (
    GenerativeFunction,
    Trace,
    Update,
)
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    Generic,
    PRNGKey,
    TypeVar,
)

# Type variables
R = TypeVar("R")
ER = TypeVar("ER", bound=EditRequest)


@Pytree.dataclass(match_args=True)
class EmptyRequest(EditRequest):
    def edit(
        self,
        key: PRNGKey,
        tr: Trace[R],
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], "EditRequest"]:
        if Diff.static_check_no_change(argdiffs):
            return tr, jnp.array(0.0), Diff.no_change(tr.get_retval()), EmptyRequest()
        else:
            request = Update(ChoiceMap.empty())
            return request.edit(key, tr, argdiffs)


@Pytree.dataclass(match_args=True)
class Regenerate(PrimitiveEditRequest):
    selection: Selection


@Pytree.dataclass(match_args=True)
class Rejuvenate(EditRequest):
    """
    The `Rejuvenate` edit request is a compositional request which utilizes
    a proposal generative function to propose a change to a trace.

    Specifying a rejuvenation requires that a user provide a `proposal`
    generative function, and an `argument_mapping`,
    which is a callable that accepts the `ChoiceMap` from the previous trace
    and produces arguments to the invocation of the generative function.

    `Rejuvenate` can be used for quick custom regeneration moves (e.g. a Gaussian at an address,
    can I propose a random walk around the previous value using another Gaussian?)
    as well as larger structured proposals.

    The logic of this move is equivalent to the logic of
    Metropolis-Hastings with a custom proposal
    (which defines the MCMC kernel of the Metropolis-Hastings algorithm)
    _without the accept-reject step_. It uses the same proposal Q
    (provided in `Rejuvenate.proposal`) for the K and L ingredients of
    the SMCP3 move. The accept-reject ratio is returned as the SMCP3 weight of the move.
    """

    proposal: GenerativeFunction[Any]
    argument_mapping: Callable[[ChoiceMap], Any] = Pytree.static()

    def edit(
        self,
        key: PRNGKey,
        tr: Trace[Any],
        argdiffs: Argdiffs,
    ) -> tuple[Trace[Any], Weight, Retdiff[Any], "EditRequest"]:
        chm = tr.get_choices()
        fwd_proposal_args = self.argument_mapping(chm)
        key, sub_key = jrand.split(key)
        proposed_change, fwd_proposal_score, _ = self.proposal.propose(
            sub_key, fwd_proposal_args
        )
        request = Update(proposed_change)
        new_tr, w, retdiff, bwd_request = request.edit(key, tr, argdiffs)
        assert isinstance(bwd_request, Update)
        bwd_chm = bwd_request.constraint
        bwd_proposal_args = self.argument_mapping(bwd_chm)
        bwd_proposal_score, _ = self.proposal.assess(bwd_chm, bwd_proposal_args)
        final_weight = w + bwd_proposal_score - fwd_proposal_score
        return (
            new_tr,
            final_weight,
            retdiff,
            Rejuvenate(self.proposal, self.argument_mapping),
        )


# NOTE: can be used in an unsafe fashion!
@Pytree.dataclass(match_args=True)
class DiffAnnotate(Generic[ER], EditRequest):
    """
    The `DiffAnnotate` request can be used to introspect on the values of type `Diff` (primal and change tangent) values flowing
    through an edit program.

    Users can provide an `argdiff_fn` and a `retdiff_fn` to manipulate changes. Note that, this introspection is inherently unsafe, users should expect:

        * If you convert `Argdiffs` in such a way that you _assert_ that a value hasn't changed (when it actually has), the edit computation will be incorrect. Similar for the `Retdiff`.
    """

    request: ER
    argdiff_fn: Callable[[Argdiffs], Argdiffs] = Pytree.static(default=lambda v: v)
    retdiff_fn: Callable[[Retdiff[Any]], Retdiff[Any]] = Pytree.static(
        default=lambda v: v
    )

    def edit(
        self,
        key: PRNGKey,
        tr: Trace[R],
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], "EditRequest"]:
        new_argdiffs = self.argdiff_fn(argdiffs)
        tr, w, retdiff, bwd_request = self.request.edit(key, tr, new_argdiffs)
        new_retdiff = self.retdiff_fn(retdiff)
        return tr, w, new_retdiff, bwd_request
